'''Wraps algorithms with SklearnAlgorithm and registers them.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List

from .sklearn_algorithm import SklearnAlgorithm
from ...algorithms.impl.algorithm import MemoryAlgorithmCatalog
from ...problem_def.task import DataType, TaskType
from ...searcher.param_range import ParamRange
from ...wrangler.constants import RangeMethod


ALGORITHMS: List[Dict[str, Any]] = [
    # Classification Algorithms
    {
        'name': 'sklearn.linear_model.LogisticRegression',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {
            'for_tests': ['true'],
            'supports_random_seed': ['true']
        },
        # Hyperparams
        'penalty': 'l2',
        'dual': False,
        'tol': 1e-4,
        'C': ParamRange(method=RangeMethod.LIST,
                        prange=[0.1, 1.0, 10.0, 100.0],
                        default=1.0),
        'fit_intercept': True,
        'intercept_scaling': 1.0,
        'class_weight': ParamRange(method=RangeMethod.LIST,
                                   prange=['balanced', None]),
        'random_state': None,
        'solver': 'lbfgs',
        'max_iter': 1000,
        'multi_class': 'auto',
        'verbose': 0,
        'warm_start': False,
        'n_jobs': None,
        'l1_ratio': None
    },
    {
        'name': 'sklearn.ensemble.RandomForestClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'n_estimators': 100,
        'criterion': 'gini',
        'max_depth': ParamRange(method=RangeMethod.LIST, prange=[8, 10, 15, None]),
        'min_samples_split': ParamRange(method=RangeMethod.LIST, prange=[2, 5, 10], default=2),
        'min_samples_leaf': 1,
        'min_weight_fraction_leaf': 0.0,
        'max_features': 'sqrt',
        'max_leaf_nodes': None,
        'min_impurity_decrease': 0.0,
        'bootstrap': True,
        'oob_score': False,
        'n_jobs': None,
        'random_state': None,
        'verbose': 0,
        'warm_start': False,
        'class_weight': ParamRange(method=RangeMethod.LIST, prange=['balanced', None]),
        'ccp_alpha': 0.0,
        'max_samples': None,
    },

    {
        'name': 'sklearn.svm.SVC',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {
            'has_predict_proba': ['false'],
            'for_tests': ['true'],
            'supports_random_seed': ['true']},
        # Hyperparams
        'C': ParamRange(method=RangeMethod.LIST,
                        prange=[0.01, 0.1, 1.0, 10.0, 100.0],
                        default=1.0),
        'kernel': 'rbf',
        'degree': 3,
        'gamma': 'scale',
        'coef0': 0.0,
        'shrinking': True,
        'probability': False,
        'tol': 1e-3,
        'cache_size': 200,
        'class_weight': ParamRange(method=RangeMethod.LIST,
                                   prange=['balanced', None]),
        'verbose': False,
        'max_iter': -1,
        'decision_function_shape': 'ovr',
        'break_ties': False,
        'random_state': None
    },
    {
        'name': 'sklearn.naive_bayes.MultinomialNB',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        # Hyperparams
        'alpha': 1.0,
        'force_alpha': True,
        'fit_prior': True,
        'class_prior': None
    },
    {
        'name': 'sklearn.discriminant_analysis.LinearDiscriminantAnalysis',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        # Hyperparams
        'solver': 'svd',
        'shrinkage': None,
        'priors': None,
        'n_components': None,
        'store_covariance': False,
        'tol': 0.0001,
        'covariance_estimator': None
    },
    {
        'name': 'sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        # Hyperparams
        'priors': None,
        'reg_param': 0.0,
        'store_covariance': False,
        'tol': 0.0001
    },
    {
        'name': 'sklearn.ensemble.AdaBoostClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'estimator': None,
        'n_estimators': 50,
        'learning_rate': 1.0,
        'random_state': None,
    },
    {
        'name': 'sklearn.svm.LinearSVC',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {
            'has_predict_proba': ['false'],
            'supports_random_seed': ['true']},
        # Hyperparams
        'penalty': 'l2',
        'loss': 'squared_hinge',
        'dual': True,
        'tol': 0.0001,
        'C': ParamRange(method=RangeMethod.LIST,
                        prange=[0.01, 0.1, 1.0, 10.0, 100.0],
                        default=1.0),
        'multi_class': 'ovr',
        'fit_intercept': True,
        'intercept_scaling': 1,
        'class_weight': ParamRange(method=RangeMethod.LIST,
                                   prange=['balanced', None]),
        'verbose': 0,
        'random_state': None,
        'max_iter': 1000,
    },
    {
        'name': 'sklearn.tree.ExtraTreeClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'criterion': 'gini',
        'splitter': 'random',
        'max_depth': ParamRange(method=RangeMethod.LIST,
                                prange=[8, 10, 15, None]),
        'min_samples_split': ParamRange(method=RangeMethod.LIST,
                                        prange=[2, 5, 10],
                                        default=2),
        'min_samples_leaf': 1,
        'min_weight_fraction_leaf': 0.0,
        'max_features': 'sqrt',
        'random_state': None,
        'max_leaf_nodes': None,
        'min_impurity_decrease': 0.0,
        'class_weight': ParamRange(method=RangeMethod.LIST, prange=['balanced', None]),
        'ccp_alpha': 0.0,
    },
    {
        'name': 'sklearn.ensemble.BaggingClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'estimator': None,
        'n_estimators': 10,
        'max_samples': 1.0,
        'max_features': 1.0,
        'bootstrap': True,
        'bootstrap_features': False,
        'oob_score': False,
        'warm_start': False,
        'n_jobs': None,
        'random_state': None,
        'verbose': 0,
    },
    {
        'name': 'sklearn.linear_model.PassiveAggressiveClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {
            'has_predict_proba': ['false'],
            'supports_random_seed': ['true']},
        # Hyperparams
        'C': 1.0,
        'fit_intercept': True,
        'max_iter': 1000,
        'tol': 0.001,
        'early_stopping': False,
        'validation_fraction': 0.1,
        'n_iter_no_change': 5,
        'shuffle': True,
        'verbose': 0,
        'loss': 'hinge',
        'n_jobs': None,
        'random_state': None,
        'warm_start': False,
        'class_weight': None,
        'average': False
    },
    {
        'name': 'sklearn.neural_network.MLPClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'hidden_layer_sizes': (100,),
        'activation': 'relu',
        'solver': 'adam',
        'alpha': 0.0001,
        'batch_size': 'auto',
        'learning_rate': 'constant',
        'learning_rate_init': 0.001,
        'power_t': 0.5,
        'max_iter': 200,
        'shuffle': True,
        'random_state': None,
        'tol': 0.0001,
        'verbose': False,
        'warm_start': False,
        'momentum': 0.9,
        'nesterovs_momentum': True,
        'early_stopping': False,
        'validation_fraction': 0.1,
        'beta_1': 0.9,
        'beta_2': 0.999,
        'epsilon': 1e-08,
        'n_iter_no_change': 10,
        'max_fun': 15000,
    },
    {
        'name': 'sklearn.ensemble.GradientBoostingClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION,
                  TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'loss': 'log_loss',
        'learning_rate': 0.1,
        'n_estimators': 100,
        'subsample': 1.0,
        'criterion': 'friedman_mse',
        'min_samples_split': ParamRange(method=RangeMethod.LIST,
                                        prange=[2, 5, 10],
                                        default=2),
        'min_samples_leaf': ParamRange(method=RangeMethod.LIST,
                                       prange=[1, 2, 5],
                                       default=1),
        'min_weight_fraction_leaf': 0.0,
        'max_depth': ParamRange(method=RangeMethod.LIST,
                                prange=[3, 5, 8, 10, 15],
                                default=3),
        'min_impurity_decrease': 0.0,
        'init': None,
        'random_state': None,
        'max_features': ParamRange(method=RangeMethod.LIST,
                                   prange=['sqrt', None]),
        'verbose': 0,
        'max_leaf_nodes': None,
        'warm_start': False,
        'validation_fraction': 0.1,
        'n_iter_no_change': None,
        'tol': 0.0001,
        'ccp_alpha': 0.0,
    },
    # Regression Algorithms
    {
        'name': 'sklearn.ensemble.GradientBoostingRegressor',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'loss': 'squared_error',
        'learning_rate': 0.1,
        'n_estimators': 100,
        'subsample': 1.0,
        'criterion': 'friedman_mse',
        'min_samples_split': ParamRange(method=RangeMethod.LIST,
                                        prange=[2, 5, 10],
                                        default=2),
        'min_samples_leaf': ParamRange(method=RangeMethod.LIST,
                                       prange=[1, 2, 5],
                                       default=1),
        'min_weight_fraction_leaf': 0.0,
        'max_depth': ParamRange(method=RangeMethod.LIST,
                                prange=[3, 5, 8, 10, 15],
                                default=3),
        'min_impurity_decrease': 0.0,
        'init': None,
        'max_features': ParamRange(method=RangeMethod.LIST,
                                   prange=['sqrt', None]),
        'alpha': 0.9,
        'verbose': 0,
        'max_leaf_nodes': None,
        'warm_start': False,
        'validation_fraction': 0.1,
        'n_iter_no_change': None,
        'tol': 1e-4,
        'ccp_alpha': 0.0,
    },
    {
        'name': 'sklearn.linear_model.LinearRegression',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'for_tests': ['true']},
        # Hyperparams
        'fit_intercept': True,
        'copy_X': True,
        'n_jobs': None,
        'positive': False,
    },
    {
        'name': 'sklearn.svm.SVR',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        # Hyperparams
        'kernel': 'rbf',
        'degree': 3,
        'gamma': 'scale',
        'coef0': 0.0,
        'tol': 1e-3,
        'C': ParamRange(method=RangeMethod.LIST,
                        prange=[0.01, 0.1, 1.0, 10.0, 100.0],
                        default=1.0),
        'epsilon': 0.1,
        'shrinking': True,
        'cache_size': 200,
        'verbose': False,
        'max_iter': -1,
    },
    {
        'name': 'sklearn.svm.LinearSVR',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'epsilon': 0.0,
        'tol': 1e-4,
        'C': ParamRange(method=RangeMethod.LIST,
                        prange=[0.01, 0.1, 1.0, 10.0, 100.0],
                        default=1.0),
        'loss': 'epsilon_insensitive',
        'fit_intercept': True,
        'intercept_scaling': 1.0,
        'dual': True,
        'verbose': 0,
        'max_iter': 1000
    },
    {
        'name': 'sklearn.ensemble.RandomForestRegressor',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {
            'for_tests': ['true'],
            'supports_random_seed': ['true']},
        # Hyperparams
        'n_estimators': 100,
        'criterion': 'squared_error',
        'max_depth': ParamRange(method=RangeMethod.LIST,
                                prange=[8, 10, 15, None]),
        'min_samples_split': ParamRange(method=RangeMethod.LIST,
                                        prange=[2, 5, 10],
                                        default=2),
        'min_samples_leaf': 1,
        'min_weight_fraction_leaf': 0.0,
        'max_features': 1.0,
        'max_leaf_nodes': None,
        'min_impurity_decrease': 0.0,
        'bootstrap': True,
        'oob_score': False,
        'n_jobs': None,
        'random_state': None,
        'verbose': 0,
        'warm_start': False,
        'ccp_alpha': 0.0,
        'max_samples': None,
    },
    {
        'name': 'sklearn.linear_model.Ridge',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'alpha': ParamRange(method=RangeMethod.LIST,
                            prange=[0.001, 0.01, 0.1, 1.0, 5.0],
                            default=1.0),
        'fit_intercept': True,
        'copy_X': True,
        'max_iter': None,
        'tol': 0.0001,
        'solver': 'auto',
        'positive': False,
        'random_state': None,
    },
    {
        'name': 'sklearn.linear_model.Lasso',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'alpha': ParamRange(method=RangeMethod.LIST,
                            prange=[0.001, 0.01, 0.1, 1.0, 5.0],
                            default=1.0),
        'fit_intercept': True,
        'precompute': False,
        'copy_X': True,
        'max_iter': 1000,
        'tol': 0.0001,
        'warm_start': False,
        'positive': False,
        'random_state': None,
        'selection': 'cyclic'
    },
    {
        'name': 'sklearn.linear_model.ElasticNet',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'alpha': 1.0,
        'l1_ratio': 0.5,
        'fit_intercept': True,
        'precompute': False,
        'max_iter': 1000,
        'copy_X': True,
        'tol': 0.0001,
        'warm_start': False,
        'positive': False,
        'random_state': None,
        'selection': 'cyclic'
    },
    {
        'name': 'sklearn.linear_model.LassoCV',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'eps': 0.001,
        'n_alphas': 100,
        'alphas': None,
        'fit_intercept': True,
        'precompute': 'auto',
        'max_iter': 1000,
        'tol': 0.0001,
        'copy_X': True,
        'cv': None,
        'verbose': False,
        'n_jobs': None,
        'positive': False,
        'random_state': None,
        'selection': 'cyclic'
    },
    {
        'name': 'sklearn.ensemble.AdaBoostRegressor',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'estimator': None,
        'n_estimators': 50,
        'learning_rate': 1.0,
        'loss': 'linear',
        'random_state': None,
    },
    {
        'name': 'sklearn.tree.ExtraTreeRegressor',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'criterion': 'squared_error',
        'splitter': 'random',
        'max_depth': ParamRange(method=RangeMethod.LIST, prange=[8, 10, 15, None]),
        'min_samples_split': ParamRange(method=RangeMethod.LIST, prange=[2, 5, 10], default=2),
        'min_samples_leaf': 1,
        'min_weight_fraction_leaf': 0.0,
        'max_features': 1.0,
        'random_state': None,
        'min_impurity_decrease': 0.0,
        'max_leaf_nodes': None,
        'ccp_alpha': 0.0
    },
    {
        'name': 'sklearn.ensemble.BaggingRegressor',
        'tasks': [TaskType.REGRESSION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'estimator': None,
        'n_estimators': 10,
        'max_samples': 1.0,
        'max_features': 1.0,
        'bootstrap': True,
        'bootstrap_features': False,
        'oob_score': False,
        'warm_start': False,
        'n_jobs': None,
        'random_state': None,
        'verbose': 0
    },
    {
        'name': 'sklearn.mixture.GaussianMixture',
        'tasks': [TaskType.DENSITY_ESTIMATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'n_components': 1,
        'covariance_type': 'full',
        'tol': 0.001,
        'reg_covar': 1e-06,
        'max_iter': 100,
        'n_init': 1,
        'init_params': 'kmeans',
        'weights_init': None,
        'means_init': None,
        'precisions_init': None,
        'random_state': None,
        'warm_start': False,
        'verbose': 0,
        'verbose_interval': 10,
    },
    {
        'name': 'sklearn.tree.DecisionTreeClassifier',
        'tasks': [TaskType.BINARY_CLASSIFICATION, TaskType.MULTICLASS_CLASSIFICATION],
        'data_types': [DataType.TABULAR],
        'tags': {'supports_random_seed': ['true']},
        # Hyperparams
        'criterion': 'gini',
        'splitter': 'best',
        'max_depth': None,
        'min_samples_split': 2,
        'min_samples_leaf': 1,
        'min_weight_fraction_leaf': 0.0,
        'max_features': None,
        'random_state': None,
        'max_leaf_nodes': None,
        'min_impurity_decrease': 0.0,
        'class_weight': None,
        'ccp_alpha': 0.0,
        'monotonic_cst': None,
    },
]


def register(catalog: MemoryAlgorithmCatalog, **kwargs):
    '''Register all the algorithms in this file.'''
    for alg in ALGORITHMS:
        new_args = alg.copy()
        new_args.update(kwargs)
        catalog.register(SklearnAlgorithm(**new_args))



'''Holds values for unbound hyperparams.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Optional

from ..wrangler.constants import RangeMethod


class ParamRange():
    '''How to pick value(s) for a hyperparam.

    :method: specifies the method for picking values (FIXED, LIST, LINEAR).
    :range: specifies value(s) for the hyperparamter.

    The typing of range depends on the value of method, and the
    type T of the hyperparameter associated with this ParamRange.

    * RangeMethod.FIXED: T
    * RangeMethod.LIST: List[T]
    * RangeMethod.LINEAR: List[T]*
        * 3 elements: [start, increment, end]
        * T must be a NumberLikeObject (supports +, < operations)
    '''
    _method: RangeMethod
    _range: Any
    _default: Any = None

    def __init__(self, method: RangeMethod, prange: Any, default: Optional[Any] = None):
        self._method = method
        self._range = prange
        self._default = default

    @property
    def method(self) -> RangeMethod:
        '''Specifies the method for picking values (FIXED, LIST, LINEAR).'''
        return self._method

    @property
    def range(self) -> Any:
        '''This is the value(s) assigned to the hyperparam.

        The typing of range depends on the value of method, and the
        type T of the hyperparameter associated with this ParamRange.

        * RangeMethod.FIXED: T
        * RangeMethod.LIST: List[T]
        * RangeMethod.LINEAR: List[T]*

            * 3 elements: [start, increment, end]
            * T must be a NumberLikeObject (supports +, < operations)
        '''
        return self._range

    @property
    def default(self) -> Any:
        '''Extract the default defaults.'''
        if self.method == RangeMethod.FIXED:
            return self._default or self.range
        return self._default

    def __str__(self) -> str:
        return f'ParamRange({self.method}: {self.range})'

    def __eq__(self, other) -> bool:
        if not isinstance(other, ParamRange):
            return False
        return (self._method == other._method
                and self._range == other._range
                and self._default == other._default)



'''Tests for splitter_auto.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...problem_def.cross_validation_config import CrossValidationConfig
from ..k_fold_splitter import KFoldSplitter

from .splitter_auto import SplitterCatalogAuto

# pylint: disable=missing-function-docstring,duplicate-code


def test_sunny_day() -> None:
    dut = SplitterCatalogAuto(cv_config=CrossValidationConfig({}))
    single_fold = dut.lookup_by_name('sklearn.model_selection.KFold')
    assert isinstance(single_fold, KFoldSplitter)


def test_lookup_by_task_defaults_properly() -> None:
    dut = SplitterCatalogAuto(cv_config=CrossValidationConfig({}))
    single_fold = dut.lookup_by_task('unknown')
    assert isinstance(single_fold, KFoldSplitter)



'''Tests for auton_mean.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring, duplicate-code, protected-access

from copy import deepcopy
import pickle
import socket
import time

import numpy as np
import pandas as pd


from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...wrangler.base_port import BasePort
from ...wrangler.dataset import Dataset
from ...wrangler.logger import Logger

from ..impl.synchronous import advance

from .auton_mean import AutonMean, AutonMeanInstance

logger = Logger(__file__).logger()

base_port = BasePort()


receiver_train = Dataset(
    covariates=pd.DataFrame({'a': [1, 2, 3], 'b': [2, 3, 4], 'c': [3, 4, 5]}))

sender_train = Dataset(
    covariates=pd.DataFrame({'a': [7, 8, 9], 'b': [8, 9, 10], 'c': [9, 10, 11]}))

empty_dataset = receiver_train.output()


LOOPBACK = '127.0.0.1'
SUNNY_DAY_PORT = base_port.next()


def test_sunny_day(assert_no_exceptions: Clearer) -> None:
    '''Sunny day with no neighbors produces the simple mean.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, SUNNY_DAY_PORT)),
                ]
            },
        },
        'my_id': 2,
    })

    alg = AutonMean()
    train = receiver_train

    dut = alg.instantiate(
        distributed=distributed_config,
    )
    assert isinstance(dut, AutonMeanInstance)

    try:
        dut.start()
        dut.fit(train)

        result_mean = dut.predict(train)
        assert result_mean is not None
        pd.testing.assert_frame_equal(
            result_mean.predictions,
            pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))
        assert_no_exceptions([dut])

    finally:
        dut.stop()


INTEGRATED_RECEIVER_PORT = base_port.next()
INTEGRATED_SENDER_PORT = base_port.next()


def test_integrated(wait_til_all_fit: Waiter,
                    assert_no_exceptions: Clearer) -> None:
    ''''Two nodes talking to each other.

    One has one set of means, the other has another set.
    We see that both nodes eventually get the same (correct) means.
    '''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, INTEGRATED_RECEIVER_PORT)),
                    ('2', (LOOPBACK, INTEGRATED_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
    }

    # TODO(Dan) Catch the case where there is only 1 class in the training data.
    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonMean()
    receiver = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause)
    )

    sender = None
    try:
        receiver.start()
        receiver.fit(receiver_train)

        check = receiver.predict(empty_dataset)

        # Confirm that the training is as expected.
        assert check is not None
        assert check.predictions is not None
        assert check.predictions.shape[0] == 1
        pd.testing.assert_frame_equal(check.predictions,
                                      pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))

        sender = alg.instantiate(distributed=DistributedConfig(distributed_sender_clause))
        sender.start()
        sender.fit(sender_train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([sender, receiver], convergence_check=True)

        receiver_got = receiver.predict(empty_dataset)
        sender_got = sender.predict(empty_dataset)

        assert receiver_got is not None
        assert sender_got is not None

        r_got_a = receiver_got.predictions['a'].iloc[0]
        r_got_b = receiver_got.predictions['b'].iloc[0]
        r_got_c = receiver_got.predictions['c'].iloc[0]

        s_got_a = sender_got.predictions['a'].iloc[0]
        s_got_b = sender_got.predictions['b'].iloc[0]
        s_got_c = sender_got.predictions['c'].iloc[0]

        assert 2.0 <= r_got_a <= 8.0
        assert 3.0 <= r_got_b <= 9.0
        assert 4.0 <= r_got_c <= 10.0

        assert 2.0 <= s_got_a <= 8.0
        assert 3.0 <= s_got_b <= 9.0
        assert 4.0 <= s_got_c <= 10.0

        assert_no_exceptions([sender, receiver])

    finally:
        receiver.stop()
        if sender is not None:
            sender.stop()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]


def test_n_same(wait_til_all_fit: Waiter,
                assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS[k - 1])) for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonMean()

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc)) for dc in distributed_clauses]

    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(receiver_train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit(duts, convergence_check=True)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(empty_dataset) for dut in duts]

        for got_prediction in got_predictions:
            assert got_prediction is not None
            pd.testing.assert_frame_equal(
                got_prediction.predictions,
                pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]


def test_telephone(wait_til_all_fit: Waiter,
                   assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS[k - 1]))
                           for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonMean()

    telephone_params = dict(alg._default_hyperparams, Lambda=1000000)

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(receiver_train)

        # Wait for the message to arrive and get processed.
        time.sleep(1)
        assert wait_til_all_fit(duts, convergence_check=True, max_time=60.0), (
            'Failed to converge in time. Check for exceptions.')

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(empty_dataset) for dut in duts]

        for got_prediction in got_predictions:
            assert got_prediction is not None
            pd.testing.assert_frame_equal(
                got_prediction.predictions,
                pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


COLS_SENDER_PORT = base_port.next()
COLS_RECEIVER_PORT = base_port.next()


def test_columns(assert_no_exceptions: Clearer) -> None:
    '''Test that columns are handled in data loading, sending and receiving.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                },
            },
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, COLS_SENDER_PORT)),
                    ('2', (LOOPBACK, COLS_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    dut = AutonMean().instantiate(distributed=distributed_config, synchronous=True)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, COLS_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    match_message = pickle.dumps((
        np.array([1, 2, 3]), ['a', 'b', 'c']
    ))

    mismatch_message = pickle.dumps((
        np.array([1, 2, 3]), ['a', 'b', 'c', 'hamster']
    ))

    try:
        dut.start()
        dut.fit(sender_train)

        advance([dut])

        sock.sendto(match_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        assert_no_exceptions([dut])

        sock.sendto(mismatch_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])

        got = dut.poll_exceptions()
        assert got is not None
        assert isinstance(got[0], NotImplementedError)
        assert 'mismatch' in str(got[1])
        assert 'hamster' in str(got[1])

        # make sure there are no other exceptions we don't expect
        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()


def test_none_values(wait_til_all_fit: Waiter,
                     assert_no_exceptions: Clearer) -> None:
    '''Auton mean can accept None values'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 2,
    })

    alg = AutonMean()
    train = Dataset(
        covariates=pd.DataFrame(
            {
                'a': [None, None, None],
                'b': [None, 2, 3],
                'c': [4, 5, 6]
            }
        )
    )

    dut = alg.instantiate(
        distributed=distributed_config,
    )
    assert isinstance(dut, AutonMeanInstance)

    try:
        dut.start()
        dut.fit(train)
        wait_til_all_fit([dut], convergence_check=False)
        result_mean = dut.predict(train)
        assert result_mean is not None
        pd.testing.assert_frame_equal(
            result_mean.predictions,
            pd.DataFrame({'a': [np.nan], 'b': [2.5], 'c': [5.0]}))

        assert_no_exceptions([dut])

    finally:
        dut.stop()


NAN_PORT = base_port.next()
NO_NAN_PORT = base_port.next()


def test_nans(assert_no_exceptions: Clearer) -> None:
    ''''Two nodes talking to each other where one has a nan.
    '''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, NAN_PORT)),
                    ('2', (LOOPBACK, NO_NAN_PORT)),
                ]
            }
        },
        'my_id': 0,
    }

    # TODO(Dan) Catch the case where there is only 1 class in the training data.
    distributed_nan_clause = deepcopy(distributed_clause)
    distributed_nan_clause['my_id'] = 1
    nan_hyperparams = {
        'Lambda': 100000,
        'omega': .7,
    }
    distributed_no_nan_clause = deepcopy(distributed_clause)
    distributed_no_nan_clause['my_id'] = 2
    no_nan_hyperparams = {
        'Lambda': 100000,
        'omega': .7,
    }

    alg = AutonMean()
    nan_node = alg.instantiate(
        distributed=DistributedConfig(distributed_nan_clause),
        synchronous=True,
        **nan_hyperparams
    )

    nan_train = Dataset(
        covariates=pd.DataFrame({'a': [np.nan, np.nan, np.nan], 'b': [8, 9, 10], 'c': [9, 10, 11]}))

    no_nan_train = Dataset(
        covariates=pd.DataFrame({'a': [7, 8, 9], 'b': [8, 9, 10], 'c': [9, 10, 11]}))

    no_nan_node = alg.instantiate(distributed=DistributedConfig(distributed_no_nan_clause),
                                  synchronous=True,
                                  **no_nan_hyperparams)
    try:
        nan_node.start()
        no_nan_node.start()

        nan_node.fit(nan_train)
        no_nan_node.fit(no_nan_train)

        # Wait for the message to arrive and get processed.
        advance([no_nan_node, nan_node])

        nan_got = nan_node.predict(empty_dataset)
        no_nan_got = no_nan_node.predict(empty_dataset)

        assert nan_got is not None
        assert no_nan_got is not None

        pd.testing.assert_frame_equal(nan_got.predictions,
                                      pd.DataFrame({'a': [8.0], 'b': [9.0], 'c': [10.0]}))

        pd.testing.assert_frame_equal(no_nan_got.predictions,
                                      pd.DataFrame({'a': [8.0], 'b': [9.0], 'c': [10.0]}))

        assert_no_exceptions([no_nan_node, nan_node])

    finally:
        nan_node.stop()
        if no_nan_node is not None:
            no_nan_node.stop()



'''An AutonML implementation of sklearn.preprocessing.OneHotEncoder.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional
from sklearn import preprocessing  # type: ignore[import]
import pandas as pd


from ....catalog.catalog import upcast
from ....problem_def.task import DataType
from ....wrangler.dataset import Dataset
from ...impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ...impl.algorithm_instance import DatasetError
from ...impl.fittable_algorithm_instance import UntrainedError
from ...impl.sklearn_like_algorithm_instance import SklearnLikeAlgorithmInstance


class OneHotModel(Algorithm):
    '''Wrapper for sklearn.preprocessing.OneHotEncoder

    ===============
    Hyperparameters
    ===============

    :categories: 'auto' or a list of array-like, default='auto'

        Categories (unique values) per feature:
            'auto':
                Determine categories automatically from the training data.
            list:
                categories\\[i\\] holds the categories expected in the ith column.
                The passed categories should not mix strings and numeric values
                within a single feature, and should be sorted in case of numeric values.

            The used categories can be found in the `categories` attribute.

    :drop: {'first', 'if_binary'} or an array-like of shape (n_features,), default=None

        Specifies a methodology to use to drop one of the categories per feature. This is
        useful in situations where perfectly collinear features cause problems, such as when
        feeding the resulting data into an unregularized linear regression model.
        However, dropping one category breaks the symmetry of the original representation and
        can therefore induce a bias in downstream models, for instance for penalized linear
        classification or regression models.

        None: retain all features (the default).
        'first': drop the first category in each feature. If only one category is present, the
        feature will be dropped entirely.
        'if_binary': drop the first category in each feature with two categories. Features with
        1 or more than 2 categories are left intact.
        array: drop[i] is the category in feature X[:, i] that should be dropped.

        When max_categories or min_frequency is configured to group infrequent categories, the
        dropping behavior is handled after the grouping.

    :dtype: number type, default=float

        Desired dtype of output.

    :handle_unknown: {'error', 'ignore', 'infrequent_if_exist'}, default='error'

        Specifies the way unknown categories are handled during transform.

        'error':
            Raise an error if an unknown category is present during transform.
        'ignore':
            When an unknown category is encountered during transform, the resulting one-hot
            encoded columns for this feature will be all zeros. In the inverse transform, an unknown
            category will be denoted as None.
        'infrequent_if_exist':
            When an unknown category is encountered during transform, the
            resulting one-hot encoded columns for this feature will map to the infrequent category
            if it exists. The infrequent category will be mapped to the last position in the
            encoding.

        During inverse transform, an unknown category will be mapped to the category denoted
        'infrequent' if it exists. If the 'infrequent' category does not exist, then transform and
        inverse_transform will handle an unknown category as with handle_unknown='ignore'.
        Infrequent categories exist based on min_frequency and max_categories.

    :min_frequency: int or float, default=None

        Specifies the minimum frequency below which a category will be considered infrequent.
        If int, categories with a smaller cardinality will be considered infrequent.
        If float, categories with a smaller cardinality than min_frequency * n_samples will be
        considered infrequent.

    :max_categories: int, default=None

        Specifies an upper limit to the number of output features for each input feature when
        considering infrequent categories. If there are infrequent categories, max_categories
        includes the category representing the infrequent categories along with the frequent
        categories. If None, there is no limit to the number of output features.
   '''
    _name = 'sklearn.preprocessing.OneHotEncoder'
    _tags = {
        'data_type': [DataType.TABULAR.name],
        'preprocessor': ['true'],
        'source': ['sklearn']
    }
    _default_hyperparams = {
        'categories': 'auto',
        'drop': None,
        'sparse_output': False,
        'dtype': float,
        'handle_unknown': 'ignore',
        'min_frequency': None,
        'max_categories': None,
    }
    _hyperparam_lookup = {
        'dtype': {
            "<float>": float,
            "<int>": int,
        }
    }

    def instantiate(self, **hyperparams) -> 'OneHotModelInstance':
        return OneHotModelInstance(parent=self, **hyperparams)


class OneHotModelInstance(SklearnLikeAlgorithmInstance):
    '''Wrapper for sklearn.preprocessing.OneHotEncoder'''
    _impl: preprocessing.OneHotEncoder
    _constructor = preprocessing.OneHotEncoder

    def __init__(self, parent: Algorithm, **kwargs):
        super().__init__(parent=parent, **kwargs)
        assert hasattr(self, '_impl')
        self._impl.set_output(transform='pandas')

    def fit(self, dataset: Optional[Dataset]) -> None:
        if dataset is None:
            raise DatasetError('attempt to fit with no data for {self.catalog_name}')
        cov_df = dataset.dataframe
        cov_cat_df = cov_df.select_dtypes(include=['bool', 'category', 'object'])
        self._impl.fit(X=cov_cat_df)
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if not self._trained:
            raise UntrainedError(f'attempt to predict before fit for {self.catalog_name}')

        if dataset is None:
            return None

        cov_df = dataset.dataframe
        cov_cat_df = cov_df.select_dtypes(include=['bool', 'category', 'object'])
        cov_rest_df = cov_df.select_dtypes(exclude=['bool', 'category', 'object'])
        output_list = [cov_rest_df]
        if not cov_cat_df.empty:
            encoded_cat_df = self._impl.transform(X=cov_cat_df)
            output_list = [cov_rest_df, encoded_cat_df]
        output_df = pd.concat(output_list, axis=1)

        retval = dataset.output()
        retval.dataframe = output_df

        return retval


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = OneHotModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Tests for communicator_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...config_components.distributed_config import DistributedConfig
from ...neighbor_manager.node_id import NodeID
from ...wrangler.constants import PACKAGE_NAME

from ..sockets_communicator import SocketsCommunicator

from .communicator import Communicator
from .communicator_auto import CommunicatorCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_COMMUNICATOR_SOURCE = '''
from queue import Queue
from typing import Any, List, Optional
from ngautonml.communicators.impl.communicator import Communicator
from ngautonml.neighbor_manager.event import Event
from ngautonml.communicators.impl.communicator_catalog import CommunicatorCatalog
from ngautonml.neighbor_manager.node_id import NodeID


class FakeCommunicator(Communicator):
    name = 'fake_communicator'
    tags = {
        'some_tag': ['some_value']
    }

    def start(self,
              queue: Queue,
              timeout: float
              ) -> None:
        pass

    def send(self,
             dest_id: NodeID,
             payload: bytes) -> int:
        pass

    def send_all(self, payload: bytes) -> int:
        pass

    def stop(self) -> None:
        pass


def register(catalog: CommunicatorCatalog):
    catalog.register(FakeCommunicator)
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'communicators' / 'fake_communicator.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_COMMUNICATOR_SOURCE)
    return retval


def test_communicator_catalog_auto() -> None:
    dut = CommunicatorCatalogAuto()

    clause = {
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [],
            },
        },
    }

    got = dut.lookup_by_name('sockets')(
        my_id=NodeID(0),
        known_neighbors=set(),
        distributed=DistributedConfig(clause=clause)
    )
    assert isinstance(got, SocketsCommunicator)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)

    dut = CommunicatorCatalogAuto()
    got = dut.lookup_by_name('fake_communicator')(
        my_id=NodeID(-1), known_neighbors=set(),
        distributed=DistributedConfig(clause={})
    )
    assert isinstance(got, Communicator)
    assert got.tags['some_tag'] == ['some_value']



""" Tests for data_loader.py"""
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Optional

import pandas as pd
import pytest

from ...config_components.dataset_config import DatasetConfig
from ...problem_def.problem_def import ProblemDefinition
from ...wrangler.dataset import Column, Dataset, DataType, RoleName, TaskType
from .data_loader import DataLoader

# Note that we disable flake8 error F841 (# noqa F841)
# because we have a local variable which is used
# but which static analysis can't tell is used.

# pylint: disable=redefined-outer-name, missing-function-docstring, duplicate-code


class FakeDataLoader(DataLoader):
    '''Fake Data Loader'''
    name: str = 'fake_data_loader'
    tags: Dict[str, List[str]] = {
        'input_format': ['fake_input_format'],
        'loaded_format': ['fake_loaded_format'],
    }

    def _load_train(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _load_test(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _load_ground_truth(self) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def _dataset(self, data: Any, **kwargs) -> Dataset:
        return Dataset(metadata=self._metadata)

    def validate(self, dataset: Optional[Dataset]) -> None:
        pass

    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        return Dataset(metadata=self._metadata)

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        return Dataset(metadata=self._metadata)


CLAUSE = {
    'dataset': {
        'config': 'fake_data_loader',
        'column_roles': {
            'target': {
                'name': 'a',
                'pos_label': 1
            }
        },
        'params': {
            'train_data': 'train_df'
        }
    },
    'problem_type': {
        'task': 'binary_classification'
    }
}


@pytest.fixture(scope='session')
def fake_data_loader() -> FakeDataLoader:
    pdef = ProblemDefinition(CLAUSE)
    train_df = pd.DataFrame()  # noqa: F841 pylint: disable=unused-variable
    dataset_config = pdef.get_conf(pdef.Keys.DATASET)
    assert isinstance(dataset_config, DatasetConfig)
    return FakeDataLoader(config=dataset_config)


def test_metadata_train(fake_data_loader: FakeDataLoader) -> None:
    got = fake_data_loader.load_train()
    assert got is not None
    assert got.metadata.roles == {RoleName.TARGET: [Column('a')]}
    assert got.metadata.pos_labels == {RoleName.TARGET: 1}
    assert got.metadata.task == TaskType.BINARY_CLASSIFICATION
    assert got.metadata.data_type == DataType.TABULAR


def test_metadata_test(fake_data_loader: FakeDataLoader) -> None:
    got = fake_data_loader.load_test()
    assert got is not None
    metadata = got.metadata
    assert metadata.roles == {RoleName.TARGET: [Column('a')]}
    assert metadata.pos_labels == {RoleName.TARGET: 1}
    assert metadata.task == TaskType.BINARY_CLASSIFICATION
    assert metadata.data_type == DataType.TABULAR



'''Implementation of Ranker'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .ranker import Ranker


class RankerImpl(Ranker):
    '''Real implementation of the Ranker'''



'''Tests for connect.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ..algorithms.impl.algorithm import InputKeyError
from .connect import ConnectorModel
from ..wrangler.dataset import Dataset
# pylint: disable=missing-function-docstring,duplicate-code


def test_sunny_day() -> None:
    model = ConnectorModel()
    instance = model.instantiate(b='a', c='b')

    dataset = Dataset(
        a=[1, 2, 3],
        b='something else'
    )

    got = instance.predict(dataset)
    want = Dataset(
        b=[1, 2, 3],
        c='something else'
    )

    assert got == want


def test_discard_key() -> None:
    model = ConnectorModel()
    instance = model.instantiate(b='a')
    dataset = Dataset(
        a=[1, 2, 3],
        x=[4, 5, 6]
    )

    got = instance.predict(dataset)
    want = Dataset(
        b=[1, 2, 3]
    )

    assert got == want


def test_key_missing() -> None:
    model = ConnectorModel()
    instance = model.instantiate(b='a', c='b')
    dataset = Dataset(
        a=[1, 2, 3]
    )

    with pytest.raises(InputKeyError):
        instance.predict(dataset)


LAYERS_DATASET = Dataset(
    base_key1='base_value1',
    base_key2='base_value2',
    base_key3='base_value3'
)


def test_layers_instance_only_selects() -> None:
    model1 = ConnectorModel(model_key1='base_key1')
    instance1 = model1.instantiate()
    got1 = instance1.predict(LAYERS_DATASET)
    assert got1 == Dataset(
        model_key1='base_value1'
    )


def test_layers_one_model_two_instances() -> None:
    model2 = ConnectorModel(model_key1='base_key1', model_key2='base_key2')
    instance2 = model2.instantiate()
    got2 = instance2.predict(LAYERS_DATASET)
    assert got2 == Dataset(
        model_key1='base_value1',
        model_key2='base_value2'
    )

    instance2a = model2.instantiate(model_key3='base_key3')
    got2a = instance2a.predict(LAYERS_DATASET)
    assert got2a == Dataset(
        model_key1='base_value1',
        model_key2='base_value2',
        model_key3='base_value3'
    )


def test_deep_connect_sunny_day() -> None:
    model = ConnectorModel()
    instance = model.instantiate(b=['a', 'aa'], c='d')

    dataset = Dataset(
        a=Dataset(
            aa=[1, 2, 3]
        ),
        d='something else'
    )

    got = instance.predict(dataset)
    want = Dataset(
        b=[1, 2, 3],
        c='something else'
    )

    assert got == want


def test_deep_dereference_fail() -> None:
    model = ConnectorModel()
    instance = model.instantiate(b=['bug', 'mismatch'], c='d')

    dataset = Dataset(
        bug=Dataset(
            mismatchedkey=[1, 2, 3]
        ),
        d='something else'
    )

    with pytest.raises(InputKeyError, match='bug.*mismatch'):
        instance.predict(dataset)


def test_deep_deep_dereference() -> None:
    model = ConnectorModel(
        target=['timeseries', 'target', 'dataframe'],
        static=['static', 'static']
    )
    instance = model.instantiate()

    dataset = Dataset(
        timeseries=Dataset(
            target=Dataset(
                dataframe=[1, 2, 3, 4]
            )
        ),
        static=Dataset(
            static=[5, 6, 7, 8]
        )
    )
    got = instance.predict(dataset)

    want = Dataset(
        target=[1, 2, 3, 4],
        static=[5, 6, 7, 8]
    )

    assert got == want



'''Tests for SimpleExecutablePipeline'''
import pickle

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import re
from typing import Dict, Optional

import numpy as np
import pytest

from ...algorithms.connect import ConnectorModel
from ...algorithms.impl.algorithm import Algorithm
from ...algorithms.impl.algorithm_instance import AlgorithmInstance
from ...algorithms.impl.fittable_algorithm_instance import UntrainedError
from ...algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator, StepDesignator
from ...instantiator.executable_pipeline import FitError
from ...wrangler.dataset import Dataset
from ..cucumber import JarOfCucumbers
from .simple_executable_pipeline import SimpleExecutablePipeline

# pylint: disable=missing-function-docstring,missing-class-docstring,protected-access,duplicate-code


class FakeFitter(Algorithm):
    '''Pretend to fit data.'''
    _name = 'fitter_model'

    def instantiate(self, **kwargs) -> AlgorithmInstance:
        return FakeFitterInstance(parent=self, **kwargs)


class FakeFitterInstance(FittableAlgorithmInstance):
    '''Pretend to fit data.'''
    _hyperparams: Dict[str, str]
    _training_data: Dataset

    def __init__(self, parent, serialized_model: Optional[bytes] = None, **kwargs):
        super().__init__(parent=parent)
        self._training_data = Dataset()
        if serialized_model is not None:
            self._training_data = pickle.loads(serialized_model)
        self._hyperparams = kwargs

    def deserialize(self, serialized_model: bytes) -> 'FakeFitterInstance':
        self._training_data = pickle.loads(serialized_model)
        return self

    def serialize(self) -> bytes:
        '''Return a serialized version of a trained model.'''
        return pickle.dumps(self._training_data, pickle.HIGHEST_PROTOCOL)

    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Fit a model based on train data.'''
        if dataset is not None:
            self._training_data = Dataset(dataset.copy())
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()
        retval.update(dataset)
        retval.update(self._hyperparams)
        return retval


REFERENCE_DATASET = Dataset(
    a_key='a_value',
    another_key='another_value',
)


def strip_serial(key: StepDesignator) -> str:
    return re.sub('[0-9]+', 'NUM', key)


def test_fit_sunny_day() -> None:
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    connector = ConnectorModel()
    fitter = FakeFitter()
    pipe.step(
        model=connector,
        a_key_renamed='a_key',
        another_key='another_key').set_name('connector')
    pipe.step(model=fitter).set_name('fitter')

    dut = SimpleExecutablePipeline(pipe)
    got_fit: JarOfCucumbers = dut.fit(dataset=REFERENCE_DATASET)

    got_stripped_jar = {strip_serial(key): val for key, val in got_fit.items()}
    assert set(got_stripped_jar.keys()) == set({
        '@connect_NUM@connector', '@fitter_model_NUM@fitter'})

    got_fitter_cucumber = got_stripped_jar['@fitter_model_NUM@fitter']
    got_fitter_instance = got_fitter_cucumber.deserialize_model(alg=FakeFitter())
    assert isinstance(got_fitter_instance, FakeFitterInstance)

    # connector will transform the data before it gets saved in the fitter
    want_fitter = {
        'a_key_renamed': 'a_value',
        'another_key': 'another_value',
    }
    assert got_fitter_instance._training_data == want_fitter


def test_predict_sunny_day() -> None:
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    connector = ConnectorModel()
    fitter = FakeFitter()
    pipe.step(model=connector,
              a_key_renamed='a_key',
              another_key='another_key').set_name('connector')
    pipe.step(model=fitter, new_key='new_value').set_name('fitter')

    dut = SimpleExecutablePipeline(pipe)
    dut.fit(dataset=REFERENCE_DATASET)

    # Connector will rename a key, FakeFitter will add a new key and value
    got_predict = dut.predict(dataset=REFERENCE_DATASET)
    want_predict = {
        'a_key_renamed': 'a_value',
        'another_key': 'another_value',
        'new_key': 'new_value'
    }
    assert got_predict.prediction is not None
    assert set(got_predict.prediction.keys()) == set(want_predict.keys())
    assert set(got_predict.prediction.values()) == set(want_predict.values())


def test_fit_must_precede_predict() -> None:
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    fitter = FakeFitter()
    pipe.step(model=fitter).set_name('fitter')
    dut = SimpleExecutablePipeline(pipeline=pipe)

    # the only model in this pipeline is the FakeFitter, which notably
    #   does *not* raise an UntrainedError if predict() is called
    #   before fit.  Thus, we test that SimpleExecutablePipeline does so
    #   natively.
    with pytest.raises(UntrainedError):
        dut.predict(dataset=Dataset())


def test_subpipeline_designator_propagation() -> None:
    '''Test that subpipelines in a parallel step are saved with the same
            designator as their parent pipeline.
    '''
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    par1 = pipe.new(name='par1')
    par1.step(model=ConnectorModel(), new_k1='a_key')
    par2 = pipe.new(name='par2')
    par2.step(model=ConnectorModel(), new_k2='another_key')
    pipe.parallel(par1=par1, par2=par2)
    pipe.step(model=ConnectorModel(),
              from_par1=['par1', 'new_k1'],
              from_par2=['par2', 'new_k2'])

    dut = SimpleExecutablePipeline(pipeline=pipe)
    assert dut.designator == Designator('foo')
    got_fit: JarOfCucumbers = dut.fit(dataset=REFERENCE_DATASET)
    for cucumber in got_fit.values():
        assert cucumber.pipeline_designator == Designator('foo')


class RandomAlgorithmInstance(AlgorithmInstance):

    def __init__(self, parent: Algorithm, **hyperparams):
        super().__init__(parent)
        if 'random_seed' in hyperparams:
            # TODO(piggy): use generator instead of setting globally to avoid flakiness
            np.random.seed([hyperparams['random_seed']])

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()
        retval['result'] = np.random.randint(0, 10, 5)
        return retval


class RandomAlgorithm(Algorithm):
    def instantiate(self, **hyperparams) -> RandomAlgorithmInstance:
        return RandomAlgorithmInstance(parent=self, **self.hyperparams(**hyperparams))


def test_set_seed() -> None:
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    pipe.step(model=RandomAlgorithm(random_seed=1234))
    dut = SimpleExecutablePipeline(pipeline=pipe)

    dut.fit(Dataset())
    got = dut.predict(Dataset())

    # For seed == 1234
    assert got.prediction is not None
    assert (got.prediction['result'] == [2, 1, 9, 4, 2]).all()


class BrokenAlg(Algorithm):

    def instantiate(self, **hyperparams) -> FittableAlgorithmInstance:
        return BrokenAlgInstance(parent=self, **hyperparams)


class BrokenAlgInstance(FittableAlgorithmInstance):
    _name = 'fit is broken'

    def fit(self, dataset: Optional[Dataset]) -> None:
        if dataset is None or 'better_data' not in dataset:
            _ = 1 / 0
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        return dataset


def test_fit_error_reset() -> None:
    '''If we get an error while fitting, the results of predict() will always
    be a FitError.

    If we refit on data that works correctly, we want this error to go away.
    '''
    pipe = BoundPipeline(name=Designator('conditionally_broken_fit'))
    pipe.step(model=BrokenAlg(name='conditionally_broken_alg'))
    dut = SimpleExecutablePipeline(pipeline=pipe)
    dut.set_fit_error(FitError())

    test_data = Dataset({'hamster': 'gerbil'})

    dut.fit(dataset=Dataset({'better_data': 'good data goes here'}))

    got_fixed = dut.predict(dataset=test_data)
    assert got_fixed.prediction == test_data



'''Catalog for metrics that can be used to evaulate pipelines'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Union

from ...catalog.memory_catalog import MemoryCatalog
from ...metrics.impl.metric import Metric
from ...problem_def.problem_def import ProblemDefinition


class MetricCatalog(MemoryCatalog[Metric], metaclass=abc.ABCMeta):
    '''Base class for metric catalogs'''

    def register(self, obj: Metric, name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)

    def lookup_metrics(self, pd: ProblemDefinition) -> Dict[str, Metric]:
        '''Get metric objects from metric catalog based on problem definition.'''
        task = pd.task
        metric_confs = pd.metric_configs

        task_type = "None"
        if task.task_type is not None:
            task_type = task.task_type.name

        # TODO(piggy): Allow hyperparams for metrics.
        if len(metric_confs) == 0:
            metrics = self.lookup_by_tag_and(task=task_type, implements_calculate='true')
        else:
            metrics = {}
            for conf in metric_confs.values():
                metric = self.lookup_by_name(conf.catalog_name)
                metrics[metric.name] = metric

        return metrics


class MetricCatalogStub(MetricCatalog):
    '''stub'''



'''Tests for metric_auto.py.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...wrangler.constants import PACKAGE_NAME
from .aggregator import Aggregator
from .aggregator_auto import AggregatorCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_AGGREGATOR_SOURCE = '''
from ngautonml.aggregators.impl.aggregator import Aggregator, AggregatorStub
from ngautonml.aggregators.impl.aggregator_catalog import AggregatorCatalog


class FakeAggregator(AggregatorStub):
    _name = 'fake_aggregator'
    _tags = {
        'some_tag': ['some_value']
    }


def register(catalog: AggregatorCatalog):
    catalog.register(FakeAggregator())
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = retval / f'.{PACKAGE_NAME}' / 'plugins' / 'aggregators' / 'fake_aggregator.py'
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_AGGREGATOR_SOURCE)
    return retval


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)
    dut = AggregatorCatalogAuto()
    got = dut.lookup_by_name('fake_aggregator')
    assert isinstance(got, Aggregator)
    assert got.tags['some_tag'] == ['some_value']



'''Makes constructors for DataframeLoaders that wrap sklearn loaders'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List

from ..impl.data_loader_catalog import DataLoaderCatalog
from .sklearn_loader_generator import SklearnLoaderGenerator


LOADERS: List[Dict[str, Any]] = [
    {
        'name': 'sklearn.datasets.make_moons',
        'tags': {
            'input_format': ['sklearn.datasets.make_moons'],
            'loaded_format': ['pandas_dataframe'],
            'source': ['sklearn'],
            'supports_random_seed': ['true'],
            'synthetic': ['true']
        }
    },
    {
        'name': 'sklearn.datasets.load_diabetes',
        'tags': {
            'input_format': ['sklearn.datasets.load_diabetes'],
            'loaded_format': ['pandas_dataframe'],
            'supports_random_seed': ['false'],
            'uses_return_X_y': ['true'],
            'uses_as_frame': ['true']
        }
    },
    {
        'name': 'sklearn.datasets.load_breast_cancer',
        'tags': {
            'input_format': ['sklearn.datasets.load_breast_cancer'],
            'loaded_format': ['pandas_dataframe'],
            'supports_random_seed': ['false'],
            'uses_return_X_y': ['true'],
            'uses_as_frame': ['true']
        }
    }

]


def register(catalog: DataLoaderCatalog):
    '''Register all the algorithms in this file.'''
    for loader_spec in LOADERS:
        catalog.register(SklearnLoaderGenerator(**loader_spec).new_class())



'''Combines pipeline templates with hyperparameter bindings to generate bound pipelines'''
import abc
from typing import Dict, List

from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ..problem_def.problem_def import ProblemDefinition
from ..templates.impl.pipeline_step import PipelineStep
from ..templates.impl.pipeline_template import PipelineTemplate
from .bound_pipeline import BoundPipeline, BoundPipelineStub
from .designator import Designator

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class Generator(metaclass=abc.ABCMeta):
    '''Base class for pipeline generators'''

    def __init__(
        self,
        algorithm_catalog: AlgorithmCatalog,
        problem_definition: ProblemDefinition,
    ):
        self._algorithm_catalog = algorithm_catalog
        self._problem_definition = problem_definition

    @abc.abstractmethod
    def generate(self, pipeline: PipelineTemplate) -> Dict[Designator, BoundPipeline]:
        '''Generate bound pipelines for one template.'''
        raise NotImplementedError

    def generate_all(self, templates: Dict[str, PipelineTemplate]
                     ) -> Dict[Designator, BoundPipeline]:
        '''Generate bound pipelines for a set of templates.'''
        retval: Dict[Designator, BoundPipeline] = {}
        for pipeline_template in templates.values():
            retval.update(self.generate(pipeline=pipeline_template))
        return retval


class GeneratorStub(Generator):
    '''stub'''
    _name: str

    def __init__(self, **kwargs):
        name = kwargs.get('name', 'default_stub_name')
        del kwargs['name']
        for arg, cls in [
            ('algorithm_catalog', AlgorithmCatalogAuto),
        ]:
            kwargs[arg] = kwargs.get(arg, cls())
        super().__init__(**kwargs)
        self._name = name

    def generate(self, pipeline: PipelineTemplate) -> Dict[Designator, BoundPipeline]:  # pylint: disable=unused-argument
        return {Designator(pipeline.name): BoundPipelineStub(self._name)}


class GeneratorImpl(Generator):
    '''The default implementation of Generator.'''

    def generate(self, pipeline: PipelineTemplate) -> Dict[Designator, BoundPipeline]:
        step0 = pipeline.steps[0]
        future_steps: Dict[Designator, List[PipelineStep]] = step0.generate(pipeline.steps[1:])
        pipelines = [BoundPipeline.build(template_name=pipeline.name, steps=steps)
                     for steps in future_steps.values()]
        return {
            bound.designator: bound for bound in pipelines
        }



'''Tests for the Saver object'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Optional

import pandas as pd
import pytest

from ..algorithms.connect import ConnectorModel
from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ..executor.cucumber import Cucumber, JarOfCucumbers
from ..generator.designator import Designator, StepDesignator
from ..instantiator.executable_pipeline import ExecutablePipelineStub
from ..instantiator.executable_pipeline import PipelineResult, PipelineResults
from ..wrangler.dataset import Dataset

from .saver import Saver
from ..problem_def.output_config import OutputConfig

# pylint: disable=missing-function-docstring,redefined-outer-name
# pylint: disable=duplicate-code,missing-class-docstring


@pytest.fixture(scope='session')
def output_config(tmp_path_factory: pytest.TempPathFactory) -> OutputConfig:
    tmp_path = tmp_path_factory.mktemp('data')
    tmp_dir = tmp_path / 'sub'
    tmp_dir.mkdir()
    config = OutputConfig(clause={
        'path': str(tmp_dir),
        'instantiations': [
            'stub_executor_kind',
            'simple'
        ],
        'file_type': 'csv'
    })
    return config


def test_save_predictions(output_config: OutputConfig) -> None:
    dut = Saver(output_config=output_config)
    prediction_df = pd.DataFrame(
        {'a': [1, 2, 3]}
    )
    prediction = Dataset()
    prediction.predictions = prediction_df
    result = PipelineResult(
        executable_pipeline=ExecutablePipelineStub(),
        prediction=prediction)
    got_path = dut.save(result=result)
    assert output_config.path is not None
    want_path = output_config.path / 'stub_pipeline'
    assert got_path == want_path

    train_pred_path = want_path / 'train_predictions.csv'
    assert train_pred_path.exists()

    got_df = pd.read_csv(str(train_pred_path))
    pd.testing.assert_frame_equal(got_df, prediction_df)


class FakeAlgorithm(Algorithm):
    _name = 'fake_algorithm'
    _default_hyperparams = {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',
    }
    _tags = {}

    def instantiate(self, **hyperparams) -> 'FakeAlgorithmInstance':
        return FakeAlgorithmInstance(parent=self, **hyperparams)


class FakeAlgorithmInstance(FittableAlgorithmInstance):
    def serialize(self) -> bytes:
        return b'fake algorithm instance model\n'

    def fit(self, dataset: Optional[Dataset]) -> None:
        _ = dataset
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        return dataset


DES1 = StepDesignator('a')
DES2 = StepDesignator('b')


class FakeExecutablePipeline(ExecutablePipelineStub):
    '''Fake trained executable pipeline containing 2 models'''

    def __init__(self):
        super().__init__(trained=True)

    def cucumberize_all(self) -> JarOfCucumbers:
        retval = JarOfCucumbers()
        alg1 = ConnectorModel(param='value')
        impl1 = alg1.instantiate()
        alg2 = FakeAlgorithm()
        impl2 = alg2.instantiate()
        retval[DES1] = Cucumber(
            impl=impl1,
            filename=DES1,
            pipeline_designator=Designator('foo'),
            hyperparams=alg1.hyperparams()
        )
        retval[DES2] = Cucumber(
            impl=impl2,
            filename=DES2,
            pipeline_designator=Designator('foo'),
            hyperparams=alg2.hyperparams()
        )
        return retval

    @property
    def locked(self) -> bool:
        return False


def test_save_all_models(output_config: OutputConfig) -> None:
    assert output_config.path is not None
    pipeline = FakeExecutablePipeline()
    results = PipelineResults({
        Designator('foo'): PipelineResult(prediction=Dataset(),
                                          executable_pipeline=pipeline)
    })

    dut = Saver(output_config=output_config)
    got = dut.save_all_models(results)

    assert got == {
        DES1: Path('a.pkl'),
        DES2: Path('b.pkl'),
    }

    file1_contents = (output_config.path / 'models' / 'foo' / got[DES1]).open(mode='rb').read()
    cucumber1 = Cucumber.deserialize(file1_contents)
    assert isinstance(cucumber1, Cucumber)
    assert cucumber1.filename == DES1
    assert cucumber1.hyperparams == {
        'param': 'value',
    }

    file2_contents = (output_config.path / 'models' / 'foo' / got[DES2]).open(mode='rb').read()
    cucumber2 = Cucumber.deserialize(file2_contents)
    assert isinstance(cucumber2, Cucumber)
    assert cucumber2.filename == DES2
    assert cucumber2.hyperparams == {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',

    }
    assert cucumber2.serialized_model == b'fake algorithm instance model\n'



'''
A model that takes a dataset and parses its columns
by data type (int, float, string, datetime)

Currently only parses object columns into categorical.
'''
# pylint: disable=duplicate-code
from typing import Optional
import pandas as pd

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset, RoleName


class ColumnParserInstance(AlgorithmInstance):
    '''Parses columns of dataset by data type'''

    def __init__(self, parent: Algorithm):
        super().__init__(parent=parent)

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        '''Parses columns of dataset by data type'''
        if dataset is None:
            return None

        retval = dataset.output()

        data = dataset.dataframe

        if RoleName.TIME in dataset.metadata.roles:
            assert len(dataset.metadata.roles[RoleName.TIME]) == 1, (
                'BUG: dataset must have exactly 1 time column in roles metadata.')
            time_col_name = dataset.metadata.roles[RoleName.TIME][0].name
            assert time_col_name is not None, (
                'BUG: time column in roles metadata must have name'
            )
            if str(time_col_name) in data.columns:
                # Note: time column also gets parsed in forecasting_splitter.py
                # TODO(Merritt): remove this redundancy once it is no longer needed
                data[time_col_name] = pd.to_datetime(data[time_col_name])

        object_cols_df = data.select_dtypes(['object'])
        data[object_cols_df.columns] = object_cols_df.apply(
            lambda x: x.astype('category'))
        retval.dataframe = data
        return retval


class ColumnParser(Algorithm):
    '''Parses columns of dataset by data type'''
    _name: str = 'Column Parser'
    _instance_constructor: type = ColumnParserInstance
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = ColumnParser(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Interface class for objects that need to use a ProblemDef.*args

Without this we would get circularity.
'''
import abc
from enum import Enum
from typing import Union

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import ConfigComponent
from .task import Task


class ProblemDefInterface(ConfigComponent, metaclass=abc.ABCMeta):
    '''Provide an interface for ProblemDef to break circularity.'''

    class Keys(AEnum):
        '''Child classes need to define this class with all their keys.'''
        TASK = 'problem_type'
        DATASET = 'dataset'
        METRICS = 'metrics'
        OUTPUT = 'output'
        HYPERPARAMS = 'hyperparams'
        CROSS_VALIDATION = 'cross_validation'
        AGGREGATION = 'aggregation'
        CONTROL = 'control'

    @abc.abstractmethod
    def get_conf(self, config_name: Union[str, Enum, AEnum]) -> ConfigComponent:
        '''This is needed to make mypy happy.

        Dataset has a ProblemDef that is typed as a ConfigComponent to avoid circularity.
        It needs to use the get_conf method of that ProblemDef
        but mypy complains if ConfigComponent does not have that method.
        '''

    @property
    @abc.abstractmethod
    def task(self) -> Task:
        '''Explains what kind of problem we're solving.

        This affects the selection of pipeline templates and models.
        '''



'''Base class for neighbor discoverers'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from queue import Queue
from typing import Dict, List

from ...config_components.distributed_config import DistributedConfig
from ...communicators.impl.communicator import Communicator
from ...communicators.stub_communicator import CommunicatorStub


class Discoverer(metaclass=abc.ABCMeta):
    '''Base class for neighbor discoverers'''
    name: str
    tags: Dict[str, List[str]]
    _config: DistributedConfig
    _communicator: Communicator

    def __init__(self,
                 config: DistributedConfig,
                 communicator: Communicator,
                 ):
        self._config = config
        self._communicator = communicator

    @property
    def communicator(self) -> Communicator:
        '''Object that listens for inbound messages'''
        return self._communicator

    @abc.abstractmethod
    def laplacian(self) -> List[List[int]]:
        '''Return the Laplacian of the network.'''

    @abc.abstractmethod
    def start(self, queue: Queue) -> None:
        '''Start up the discover.

        The queue is our link to the NeighborManager.
        '''

    def stop(self) -> None:
        '''Stop all supporting threads.'''
        # The default implementation only stops communicator threads.
        self._communicator.stop()


class DiscovererStub(Discoverer):
    '''Stub Discoverer'''
    name = 'stub_discoverer'
    tags = {}

    def __init__(self):
        super().__init__(
            config=DistributedConfig({}),
            communicator=CommunicatorStub())

    def start(self, queue: Queue) -> None:
        pass

    def laplacian(self) -> List[List[int]]:
        return []



'''Build a shelf of catalogs based just on paths.'''

from pathlib import Path
from typing import List, TypeVar

from .memory_catalog import MemoryCatalog
from .catalog_shelf import CatalogShelf


T = TypeVar("T")


class PathedCatalog(CatalogShelf[T]):
    '''Make a shelf of memory-based catalogs based on a path.'''

    def __init__(self, paths: List[Path], **kwargs):
        results: List[MemoryCatalog[T]] = []
        for path in paths:
            catalog = MemoryCatalog[T](**kwargs)
            catalog.load(module_directory=path)
            results.append(catalog)
        super().__init__(*results)



'''Tests for auton_gaussian_mixture_models.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,too-many-locals

from copy import deepcopy
import pickle
import socket
from typing import Any, Dict, Optional, Tuple, List

import numpy as np
import pandas as pd
import pytest

from sklearn import datasets  # type: ignore[import]

from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...wrangler.base_port import BasePort
from ...wrangler.constants import Defaults
from ...wrangler.dataset import Dataset, DatasetKeys, Metadata
from ...wrangler.logger import Logger

from ..impl.algorithm_auto import AlgorithmCatalogAuto
from ..impl.synchronous import advance

from .GMM import GMM  # type: ignore
from .auton_gaussian_mixture_models import (
    AutonGaussianMixtureModels,
    AutonGaussianMixtureModelsInstance,
    AutonGaussianMixtureModelsNeighbor)

logger = Logger(__file__, to_stdout=False).logger()


DIST_PARAMS = {
    'K': 2,
    'Lambda': 10.0,
    'omega': .6667,
    'full_covariance': True
}


def load_simulated_data(mix: np.ndarray,
                        means: np.ndarray,
                        covs: np.ndarray,
                        n_points: int,
                        random_state: Optional[np.random.RandomState] = None  # pylint: disable=no-member,line-too-long
                        ) -> Dataset:
    '''
    Simulate data produced by a gaussian mixture.

    Parameters:
    mix: probabilities of each gaussian in the mixture, sums to 1
    means: means of each gaussian in the mixture
    covs: covariances of each gaussian in the mixture
    n_points: number of points to sample
    '''
    random_state = random_state or np.random.RandomState(1)  # pylint: disable=no-member,line-too-long
    assert sum(mix) == 1
    choices = random_state.choice(len(mix), size=n_points, p=mix)
    data = np.array([
        random_state.multivariate_normal(means[k], covs[k])
        for k in choices])

    return Dataset(
        metadata=Metadata(),
        dataframe=pd.DataFrame(data)
    )


def load_regression_dataset() -> Tuple[Dataset, Dataset, Dataset]:
    # Load the diabetes dataset
    diabetes_x, _ = datasets.load_diabetes(return_X_y=True)

    # Use only one feature
    # diabetes_x = diabetes_x[:, np.newaxis, 2]

    # Split the data into training/testing sets
    diabetes_train = diabetes_x[:-20]
    diabetes_test = diabetes_x[-20:]

    columns = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
    metadata = Metadata()

    dataset_train = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame(diabetes_train, columns=columns)
    )
    dataset_test = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame(diabetes_test, columns=columns)
    )

    reduced_train = Dataset(
        metadata=metadata,
        dataframe=pd.DataFrame(diabetes_train[:2], columns=columns)
    )
    return (dataset_train, dataset_test, reduced_train)


def build_clause(adjacency: Dict[str, List[int]],
                 nodes_and_endpoints: List[Tuple[str, Tuple[str, int]]],
                 my_id: int) -> Dict[str, Any]:
    return {
        'polling_interval': '0.1',

        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': my_id,
    }


base_port = BasePort()
LOOPBACK = '127.0.0.1'


def norm2_diff_threshold(dataset: Dataset,
                         port: int,
                         k_: int = 2,
                         full_covariance: bool = False,
                         random_seed: int = 1701) -> float:
    '''Return a maximum reasonable value for norm2_diff between two nodes.

    For GMMs, the meaning of the parameters depends on the scale of the
        data; thus the meaning of norm2_diff is also dependent on scale.

    In this function, we run a GMM on the given dataset on one node,
        perturb the parameters of the resulting mixture by a 'reasonable amount'
        as determined by heuristics, and then compute the norm2_diff between the
        perturbed and non-perturbed models.

    Assumes nodes were trained on data from the same distribution, but not
        neccesarily the exact same data.

    k_: number of gaussians
    full_covariance: whether we are using a full cov matrix

    TODO(Merritt): improve this expanation after talking to Kyle or Dan
    '''
    # pylint: disable=too-many-locals
    random_state = np.random.RandomState(random_seed)  # pylint: disable=no-member,line-too-long
    distributed_config = DistributedConfig(clause=build_clause(
        adjacency={'2': []}, nodes_and_endpoints=[('2', (LOOPBACK, port))], my_id=2))

    data = dataset.dataframe.to_numpy()
    n_rows = data.shape[0]
    std = np.std(data, axis=0)
    q75, q25 = np.quantile(data, [0.75, 0.25], axis=0)
    iqr = q75 - q25
    h = 0.9 * np.minimum(std, iqr / 1.34) * n_rows ** (-1.0 / 5.0)

    reference = AutonGaussianMixtureModels().instantiate(
        distributed=distributed_config,
        random_seed=1701,
        K=k_,
        Lambda=0.0,
        omega=0.6667,
        full_covariance=full_covariance)

    assert isinstance(reference, AutonGaussianMixtureModelsInstance)

    try:
        reference.start()
        reference.fit(dataset)
        state = reference.my_state
        assert state is not None

        fudge_h = h * 2  # Multiply by 2 to make tests more permissive
        mix_hat = np.clip(state.mix + random_state.random(state.mix.size) * 0.01, 0, 1)
        means_hat = state.means + random_state.normal(size=state.means.shape) / 1.96 * fudge_h
        if not state.full_covariance:
            cov_tril_hat = np.clip(state.cov_tril
                                   + (np.random.normal(size=state.cov_tril.shape)
                                      / 1.96 * fudge_h**2),
                                   1e-2 * fudge_h**2,
                                   np.inf)
        else:
            cov_tril_hat = (cov_tril_hat
                            + np.random.normal(size=state.cov_tril.shape) / 1.96 * fudge_h)

        state_hat = state.__class__(
            K=state.K,
            full=state.full_covariance,
            mix=mix_hat,
            means=means_hat,
            cov_tril=cov_tril_hat)

        retval = np.sqrt(state.norm2_diff(state_hat))
        logger.debug("threshold: %s", retval)
        return retval
    finally:
        reference.stop()


def test_should_send() -> None:
    '''_should_send returns True iff our model changed.'''

    config = DistributedConfig({})
    dut = AutonGaussianMixtureModelsNeighbor(
        mix=np.asarray([1, 2]),
        means=np.asarray([[1, 2], [3, 4]]),
        cov_tril=np.asarray([[1, 2], [3, 4]]),
        full=False
    )
    same_as_dut = deepcopy(dut)
    different_from_dut = AutonGaussianMixtureModelsNeighbor(
        mix=np.asarray([10, 20]),
        means=np.asarray([[10, 20], [30, 40]]),
        cov_tril=np.asarray([[10, 20], [30, 40]]),
        full=False
    )

    assert dut.state_differs(
        distributed=config,
        other=different_from_dut
    ) is True

    assert dut.state_differs(
        distributed=config,
        other=same_as_dut
    ) is False


SUNNY_DAY_PDF_FULL_PORT = base_port.next()


def test_sunny_day_pdf_full(assert_no_exceptions: Clearer) -> None:
    '''Compare mean log PDF of our GMM to sklearn's GMM on one node, with full covariance.'''
    distributed_config = DistributedConfig(
        clause=build_clause(adjacency={'2': []},
                            nodes_and_endpoints=[('2', (LOOPBACK, SUNNY_DAY_PDF_FULL_PORT))],
                            my_id=2))

    _k = 3
    _lambda = 1.0
    full_cov = True
    omega = 0.667

    config = {'K': _k,
              'covariance': 'full' if full_cov else 'adsf',
              'lambda': _lambda,
              'omega': omega}
    kyle_model = GMM(config)  # TODO(Piggy): remove reference to Kyle's code

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.mixture.GaussianMixture')

    train, test, _ = load_regression_dataset()
    sklearn_train = train.output()
    sklearn_train[DatasetKeys.COVARIATES.value] = train.dataframe

    alg = AutonGaussianMixtureModels()

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's gaussian mixture, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        Lambda=_lambda, K=_k, omega=omega, random_seed=1701,
        full_covariance=full_cov)
    assert isinstance(dut, AutonGaussianMixtureModelsInstance)

    sklearn_dut = sklearn_alg.instantiate(
        n_components=3, covariance_type='full', random_state=1701
    )

    try:
        dut.start()

        sklearn_dut.fit(sklearn_train)
        kyle_model.fit(train.dataframe.to_numpy(), [])
        dut.fit(train)

        got = dut.mean_log_pdf(test)
        kyle_got = kyle_model.lp(test.dataframe.to_numpy()) / test.dataframe.shape[0]
        sklearn_got = sklearn_dut._impl.score(test.dataframe)  # pylint: disable=protected-access

        assert kyle_got == pytest.approx(sklearn_got, abs=0.5)
        assert got == pytest.approx(sklearn_got, abs=0.5)
        assert got == pytest.approx(kyle_got, abs=0.000001)
        assert_no_exceptions([dut])
    finally:
        dut.stop()


SUNNY_DAY_SAMPLE_FULL_PORT = base_port.next()


# TODO(Merritt/Piggy): this is xpassing, but really slow.
# not sure why we're asserting sum(mix) == 1, but it's being violated.
# >       assert sum(state.mix) == 1
# ngautonml/algorithms/distributed/auton_gaussian_mixture_models.py:628: AssertionError
def test_sunny_day_sample_full(assert_no_exceptions: Clearer) -> None:
    '''Compare a sample from our GMM to sklearn's GMM on one node, with full covariance.'''
    distributed_config = DistributedConfig(
        clause=build_clause(adjacency={'2': []},
                            nodes_and_endpoints=[('2', (LOOPBACK, SUNNY_DAY_SAMPLE_FULL_PORT))],
                            my_id=2))
    _k = 3
    _lambda = 1.0
    full_cov = True
    omega = 0.667

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.mixture.GaussianMixture')

    train, _, _ = load_regression_dataset()
    sklearn_train = train.output()
    sklearn_train[DatasetKeys.COVARIATES.value] = train.dataframe

    alg = AutonGaussianMixtureModels()

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's GMM, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        synchronous=True,
        Lambda=_lambda, K=_k, omega=omega, random_seed=1701,
        full_covariance=full_cov)
    assert isinstance(dut, AutonGaussianMixtureModelsInstance)

    sklearn_dut = sklearn_alg.instantiate(
        n_components=3, covariance_type='full', random_state=1701
    )

    try:
        dut.start()

        sklearn_dut.fit(sklearn_train)
        dut.fit(train)

        advance([dut])

        # Test that empirical and sampled means are equal up to 2 standard deviations
        got = dut.sample(1000)
        sklearn_want, _ = sklearn_dut._impl.sample(1000)  # pylint: disable=protected-access
        got_means = got.dataframe.mean(axis=0).to_numpy()
        sklearn_means = sklearn_want.mean(axis=0)
        empirical_means = train.dataframe.mean(axis=0).to_numpy()
        empirical_sd = train.dataframe.std(axis=0).to_numpy()

        got_z_score = (got_means - empirical_means) / (empirical_sd)
        sklearn_z_score = (sklearn_means - empirical_means) / (empirical_sd)

        ninety_fifth_percentile = 1.96 / np.sqrt(1000)
        assert all(abs(got_z_score) < ninety_fifth_percentile)
        assert all(abs(sklearn_z_score) < ninety_fifth_percentile)
        assert_no_exceptions([dut])

    finally:
        dut.stop()


SUNNY_DAY_PORT_SHARED = base_port.next()


@pytest.mark.xfail(reason="Matches Kyles code but not sklearn")
def test_sunny_day_shared() -> None:
    distributed_config = DistributedConfig(clause=build_clause(
        adjacency={'2': []},
        nodes_and_endpoints=[('2', (LOOPBACK, SUNNY_DAY_PORT_SHARED))],
        my_id=2)
    )

    _k = 3
    _lambda = 1.0
    full_cov = False
    str_full_cov = 'full' if full_cov else 'tied'
    omega = 0.667

    config = {'K': _k, 'covariance': str_full_cov, 'lambda': _lambda, 'omega': omega}
    kyle_model = GMM(config)

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.mixture.GaussianMixture')

    train, test, _ = load_regression_dataset()
    sklearn_train = train.output()
    sklearn_train[DatasetKeys.COVARIATES.value] = train.dataframe

    alg = AutonGaussianMixtureModels()

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's logistic regression, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        Lambda=_lambda, K=_k, omega=omega, random_seed=1701,
        full_covariance=full_cov)
    assert isinstance(dut, AutonGaussianMixtureModelsInstance)

    sklearn_dut = sklearn_alg.instantiate(
        n_components=3, covariance_type=str_full_cov, random_state=1701
    )

    try:
        dut.start()

        sklearn_dut.fit(sklearn_train)
        kyle_model.fit(train.dataframe.to_numpy(), [])
        dut.fit(train)

        got = dut.mean_log_pdf(test)
        kyle_got = kyle_model.lp(test.dataframe.to_numpy()) / test.dataframe.shape[0]
        sklearn_got = sklearn_dut._impl.score(test.dataframe)  # pylint: disable=protected-access

        assert got == pytest.approx(kyle_got, abs=0.000001)
        assert kyle_got == pytest.approx(sklearn_got, abs=0.5)
        assert got == pytest.approx(sklearn_got, abs=0.5)
    finally:
        dut.stop()


INTEGRATED_RECEIVER_PORT = base_port.next()
INTEGRATED_SENDER_PORT = base_port.next()


@pytest.mark.xfail(reason='dut1 does not seem to learn from dut2.')
def test_integrated(wait_til_all_fit: Waiter) -> None:
    # pylint: disable=too-many-locals
    distributed_clause = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, INTEGRATED_RECEIVER_PORT)),
            ('2', (LOOPBACK, INTEGRATED_SENDER_PORT)),
        ],
        my_id=0
    )

    mean_1 = [-5, 2]
    mean_2 = [3, -1]
    cov_1 = [[1, 2], [2, 1]]
    cov_2 = [[3, -1], [-1, 3]]

    data_1 = load_simulated_data(
        mix=np.array([1]),
        means=np.array([mean_1]),
        covs=np.array([cov_1]),
        n_points=200)
    data_2 = load_simulated_data(
        mix=np.array([1]),
        means=np.array([mean_2]),
        covs=np.array([cov_2]),
        n_points=800)
    data_gmm = load_simulated_data(
        mix=np.array([.2, .8]),
        means=np.array([mean_1, mean_2]),
        covs=np.array([cov_1, cov_2]),
        n_points=1000)
    # Likelihood should be higher for a dataset with correct mixture of identical gaussians.
    data_gmm_scrambled = load_simulated_data(
        mix=np.array([.8, .2]),
        means=np.array([mean_1, mean_2]),
        covs=np.array([cov_1, cov_2]),
        n_points=1000)

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonGaussianMixtureModels()
    dut_1 = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        **DIST_PARAMS)

    assert isinstance(dut_1, AutonGaussianMixtureModelsInstance)

    dut_2 = None
    try:
        dut_1.start()
        dut_1.fit(data_1)

        dut_1_got_1_init = dut_1.mean_log_pdf(data_1)
        dut_1_got_2_init = dut_1.mean_log_pdf(data_2)
        dut_1_got_both_init = dut_1.mean_log_pdf(data_gmm)

        assert dut_1_got_1_init > dut_1_got_2_init
        assert dut_1_got_1_init > dut_1_got_both_init

        dut_2 = alg.instantiate(distributed=DistributedConfig(distributed_sender_clause),
                                **DIST_PARAMS)
        assert isinstance(dut_2, AutonGaussianMixtureModelsInstance)
        dut_2.start()
        dut_2.fit(data_2)

        wait_til_all_fit([dut_2, dut_1], convergence_check=True)

        # dut_1_got_1 = dut_1.mean_log_pdf(data_1)
        dut_1_got_2 = dut_1.mean_log_pdf(data_2)
        dut_1_got_both = dut_1.mean_log_pdf(data_gmm)
        dut_1_got_both_scrambled = dut_1.mean_log_pdf(data_gmm_scrambled)

        dut_2_got_2 = dut_2.mean_log_pdf(data_2)
        dut_2_got_1 = dut_2.mean_log_pdf(data_1)
        dut_2_got_both = dut_2.mean_log_pdf(data_gmm)
        dut_2_got_both_scrambled = dut_2.mean_log_pdf(data_gmm_scrambled)

        # dut 1 learned from dut 2

        # I do not think that dut_1 should have a better result after
        # learning from dut_2. If anything, it should be slightly worse.
        # In practice these values are super close.
        # assert dut_1_got_1 < dut_1_got_1_init

        assert dut_1_got_2 > dut_1_got_2_init
        assert dut_1_got_both > dut_1_got_both_init

        assert dut_2_got_2 > dut_2_got_1

        # dut 1 learned the correct mixture
        # assert dut_1_got_both > dut_1_got_both_scrambled

        # dut 2 learned the correct mixture
        assert dut_1_got_both + dut_2_got_both > dut_1_got_both_scrambled + dut_2_got_both_scrambled

    finally:
        dut_1.stop()
        if dut_2 is not None:
            dut_2.stop()


TWO_HALF_RECEIVER_PORT = base_port.next()
TWO_HALF_SENDER_PORT = base_port.next()
TWO_HALF_REFERENCE_PORT = base_port.next()


def test_two_half_hard_mode(assert_no_exceptions: Clearer) -> None:
    '''Two nodes with half the training behaves as (reasonably) expected.

    Hard mode means that we have completely dividided the dataset along
    a random hyperplane, so that the two halves are completely different.
    '''
    distributed_clause = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, TWO_HALF_RECEIVER_PORT)),
            ('2', (LOOPBACK, TWO_HALF_SENDER_PORT)),
        ],
        my_id=0
    )
    train, _, _ = load_regression_dataset()

    # Split the training data such that each node sees a differently distributed half of the
    # input space.
    # Kyle's method to choose a random hyperplane and use it to divide data:
    # 1. Sample a random d-dimensional vector from a standard gaussian distribution.
    # 2. Project all of our data onto that vector - dot product of that vector with data matrix
    # gives you 1 number per row in the dataset
    # 3. sort based on that number and divide data in half
    input_df = train.dataframe
    threadsafe_random_state = np.random.RandomState(seed=Defaults.SEED)  # pylint: disable=no-member,line-too-long
    random_hyperplane = threadsafe_random_state.normal(size=10)
    projection = np.dot(random_hyperplane, input_df.to_numpy().transpose())
    input_df['projection'] = projection
    input_df.sort_values(by='projection')

    input_df.drop(columns=['projection'], inplace=True)

    half_n = int(input_df.shape[0] / 2)

    train_first = Dataset(
        metadata=train.metadata,
        dataframe=input_df.iloc[:half_n]
    )
    train_second = Dataset(
        metadata=train.metadata,
        dataframe=input_df.iloc[half_n:]
    )

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    dist_params_first = dict(DIST_PARAMS)
    dist_params_first['full_covariance'] = False
    dist_params_second = dict(dist_params_first)
    dist_params_second['full_covariance'] = False

    alg = AutonGaussianMixtureModels()
    dut_first = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **dist_params_first)
    assert isinstance(dut_first, AutonGaussianMixtureModelsInstance)
    dut_second = alg.instantiate(
        distributed=DistributedConfig(distributed_sender_clause),
        synchronous=True,
        **dist_params_second)
    assert isinstance(dut_second, AutonGaussianMixtureModelsInstance)

    try:
        dut_first.start()
        dut_second.start()

        dut_first.fit(train_first)
        dut_second.fit(train_second)

        # Wait for the message to arrive and get processed.
        for _ in range(20):
            advance([dut_first, dut_second])

        # Confirm that the two nodes have similar training and thus shared information.
        got = np.sqrt(dut_second.norm2_diff(dut_first))

        logger.debug("dut_second._my_state: %s", dut_second.my_state)
        logger.debug("dut_first._my_state: %s", dut_first.my_state)

        # 1701: got: 8501.204773389072, thold: 7000.690242561011
        # 223: got: 8507.35572380122, thold: 5651.2776904352995
        # 1024:  got: 8499.591223770127, thold: 4365.575976761985
        # 1701 (h scaled down): got: 8496.921338822653, thold: 299.18100884306773
        # 1701 (h doubled): got: 8506.648547718962, thold: 11779.6298575865
        thold = norm2_diff_threshold(
            train, port=TWO_HALF_REFERENCE_PORT, k_=2, full_covariance=False, random_seed=1701)

        # raise NotImplementedError(f"got: {got}, thold: {thold}")
        # TODO(Kyle): Confirm that a threshold over 9000 is acceptable.
        assert got == pytest.approx(0, abs=thold), (
            'got: {got} is not within {thold} of 0')
        assert_no_exceptions([dut_first, dut_second])

    finally:
        dut_first.stop()
        dut_second.stop()


TWO_SAME_RECEIVER_PORT = base_port.next()
TWO_SAME_SENDER_PORT = base_port.next()


def test_two_same(wait_til_all_fit: Waiter,
                  assert_no_exceptions: Clearer) -> None:
    '''Two nodes with the same training don't change their coefficients.'''
    distributed_clause = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, TWO_SAME_RECEIVER_PORT)),
            ('2', (LOOPBACK, TWO_SAME_SENDER_PORT)),
        ],
        my_id=0)
    my_params = DIST_PARAMS.copy()
    my_params['full_covariance'] = False

    train, _, _ = load_regression_dataset()

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonGaussianMixtureModels()
    dut_receiver = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        **my_params)
    assert isinstance(dut_receiver, AutonGaussianMixtureModelsInstance)

    dut_sender = None
    try:
        dut_receiver.start()
        dut_receiver.fit(train)

        assert wait_til_all_fit([dut_receiver], convergence_check=False), (
            'Timeout while waiting for fit'
        )

        dut_sender = alg.instantiate(
            distributed=DistributedConfig(distributed_sender_clause),
            **my_params)
        assert isinstance(dut_sender, AutonGaussianMixtureModelsInstance)

        dut_sender.start()
        dut_sender.fit(train)

        # Wait for the message to arrive and get processed.
        assert wait_til_all_fit([dut_receiver, dut_sender], convergence_check=True), (
            'Timeout while waiting for convergence'
        )

        # Confirm that neighbor optimization doesn't mess up result.
        got = np.sqrt(dut_receiver.norm2_diff(dut_sender))

        # These two lines are to confirm that norm2_diff works properly
        # TODO(Merritt): move to own test case
        assert dut_receiver.norm2_diff(dut_receiver) == pytest.approx(0, abs=0.1), (
            'receiver norm2_diff with itself is not 0'
        )
        assert dut_sender.norm2_diff(dut_sender) == pytest.approx(0, abs=0.1), (
            'sender norm2_diff with itself is not 0'
        )

        logger.debug("dut_receiver._my_state: %s", dut_receiver.my_state)
        logger.debug("dut_sender._my_state: %s", dut_sender.my_state)

        assert got == pytest.approx(0, abs=15), (
            f'got: {got} is not within 15 of 0'
        )
        assert_no_exceptions([dut_sender, dut_receiver])

    finally:
        dut_receiver.stop()
        if dut_sender is not None:
            dut_sender.stop()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]
TEST_N_SAME_REFERENCE_PORT = base_port.next()


def test_n_same(wait_til_all_fit: Waiter,
                assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS[k - 1])) for k in range(1, n + 1)]
    distributed_clause = build_clause(
        adjacency=adjacency, nodes_and_endpoints=nodes_and_endpoints, my_id=0
    )
    my_params: dict[str, Any] = DIST_PARAMS.copy()
    my_params['full_covariance'] = False

    train, _, _ = load_regression_dataset()
    norm2_max = norm2_diff_threshold(
        dataset=train,
        port=TEST_N_SAME_REFERENCE_PORT,
        k_=int(my_params['K']),
        full_covariance=my_params['full_covariance'])

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonGaussianMixtureModels()

    duts: List[AutonGaussianMixtureModelsInstance] = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **my_params) for dc in distributed_clauses]
    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit(duts, convergence_check=True)

        # Confirm that neighbor optimization doesn't mess up result.

        dut0_state = duts[0].my_state_copy
        assert dut0_state is not None

        got = []
        for duti in duts[1:]:
            duti_state = duti.my_state_copy
            assert duti_state is not None
            got.append(np.sqrt(dut0_state.norm2_diff(duti_state)))
        assert all(got_diff < norm2_max for got_diff in got)
        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]
TEST_TELEPHONE_REFERENCE_PORT = base_port.next()


def test_telephone(assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS[k - 1]))
                           for k in range(1, n + 1)]
    distributed_clause = build_clause(
        adjacency=adjacency, nodes_and_endpoints=nodes_and_endpoints, my_id=0
    )

    train, _, _ = load_regression_dataset()

    thold = norm2_diff_threshold(
        train, port=TEST_TELEPHONE_REFERENCE_PORT, k_=2, full_covariance=False, random_seed=1701)

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonGaussianMixtureModels()

    telephone_params = dict(DIST_PARAMS, Lambda=1.0, omega=0.5)

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts: List[AutonGaussianMixtureModelsInstance] = [alg.instantiate(
        distributed=DistributedConfig(dc),
        synchronous=True,
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(train)

        # Wait for the message to arrive and get processed.
        for _ in range(n + 1):
            advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        dut0 = duts[0]
        got: List[np.float64] = []
        for duti in duts[1:]:
            got.append(np.sqrt(dut0.norm2_diff(duti)))

        want = [np.float64(0.0)] * len(got)

        np.testing.assert_allclose(actual=got, desired=want, atol=thold)
        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


COLS_SENDER_PORT = base_port.next()
COLS_RECEIVER_PORT = base_port.next()


def test_columns(assert_no_exceptions: Clearer) -> None:
    '''Test that columns are handled in data loading, sending and receiving.'''
    distributed_config = DistributedConfig(clause=build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, COLS_SENDER_PORT)),
            ('2', (LOOPBACK, COLS_RECEIVER_PORT)),
        ],
        my_id=2
    ))

    mean_1 = [-5, 2]
    mean_2 = [3, -1]
    cov_1 = [[1, 2], [2, 1]]
    cov_2 = [[3, -1], [-1, 3]]

    data_gmm = load_simulated_data(
        mix=np.array([.2, .8]),
        means=np.array([mean_1, mean_2]),
        covs=np.array([cov_1, cov_2]),
        n_points=1000)

    dut = AutonGaussianMixtureModels().instantiate(
        distributed=distributed_config, synchronous=True, **DIST_PARAMS)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, COLS_SENDER_PORT))

    valid_mix = np.array([0.8, 0.2])
    valid_means = np.array([[1, 2], [3, 4]])
    valid_cov_tril = np.array([[[1, 2], [3, 4]], [[1, 2], [3, 4]]])
    match_message = pickle.dumps((
        valid_mix, valid_means, valid_cov_tril, data_gmm.dataframe.columns.to_list()
    ))

    mismatch_message = pickle.dumps((
        valid_mix, valid_means, valid_cov_tril, data_gmm.dataframe.columns.to_list() + ['hamster']
    ))

    try:
        dut.start()
        dut.fit(data_gmm)

        advance([dut])

        sock.sendto(match_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        assert_no_exceptions([dut])

        sock.sendto(mismatch_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])

        got = dut.poll_exceptions()
        assert got is not None
        assert isinstance(got[0], NotImplementedError)
        assert 'mismatch' in str(got[1])
        assert 'hamster' in str(got[1])

        # make sure there are no other exceptions we don't expect
        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()



'''Tests for neighbor_manager.py'''

import os
from filelock import FileLock
import mlflow  # type: ignore

# pylint: disable=missing-function-docstring,duplicate-code

from .event import Recv
from .neighbor_manager import NeighborManagerStub
from .node_id import NodeID

LOCK_PATH = f'/var/tmp/neighbor-manager-mlflow-{os.environ.get("USER")}.lock'


def test_send_all_logging() -> None:
    with FileLock(LOCK_PATH, timeout=90):
        dut = NeighborManagerStub()
        with mlflow.start_run(nested=True) as r:
            dut.send_all(b'lorem ipsum')
            run = r
    assert run is not None

    client = mlflow.tracking.MlflowClient()

    metrics = client.get_metric_history(run_id=run.info.run_id, key='PayloadSize')
    assert len(metrics) == 1
    assert metrics[0].value == len(b'lorem ipsum')


def test_send_logging() -> None:
    with FileLock(LOCK_PATH, timeout=90):
        dut = NeighborManagerStub()
        with mlflow.start_run(nested=True) as r:
            dut.send([NodeID(1)], b'lorem ipsum')
            run = r
    assert run is not None

    client = mlflow.tracking.MlflowClient()

    metrics = client.get_metric_history(run_id=run.info.run_id, key='PayloadSize')
    assert len(metrics) == 1
    assert metrics[0].value == len(b'lorem ipsum')


def test_poll_for_events():
    with FileLock(LOCK_PATH, timeout=90):
        dut = NeighborManagerStub()
        with mlflow.start_run(nested=True) as r:
            events = dut.poll_for_events()
            run = r
        assert events == [Recv(NodeID(1), b'hello, world!')]

    assert run is not None

    client = mlflow.tracking.MlflowClient()
    recv_len = client.get_metric_history(run_id=run.info.run_id, key='RecvLen')
    assert recv_len[0].value == float(len('hello, world!'))
    recv_neighbor = client.get_metric_history(run_id=run.info.run_id, key='RecvNeighbor')
    assert recv_neighbor[0].value == float(1)



'''Tests for self_distance.py'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping


from ..algorithms.impl.neighbor_metadata import NeighborMetadata, NeighborMetadataInterface
from ..algorithms.impl.toy_neighbor_state import ToyNeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto


def test_catalog() -> None:
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'self_distance': {
                    'threshold': 0.2,  # Default is 0.5.
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('self_distance')(
        config=distributed.decider.self_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata()
    near_state = ToyNeighborState(center=0.1)
    below_state = ToyNeighborState(center=0.3)
    distant_state = ToyNeighborState(center=1.0)

    neighbor_id_ignored = NodeID(-1)

    # No previous or current state should not send.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is False

    # Previous state sent is None, but the current state is not; we should send.
    my_meta.current_state = my_state
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is True
    my_meta.sent_current_state()

    # Previous state sent and current state are the same; we should not send.
    my_meta.current_state = my_state
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is False

    # Previous state sent is center=0.0, current state is center=0.1, threshold=0.2;
    # we should not send.
    my_meta.current_state = near_state
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is False

    # Previous state sent is center=0.0, current state is center=0.3, threshold=0.2;
    # we should send.
    my_meta.current_state = below_state
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is True
    my_meta.sent_current_state()

    # Previous state sent is center=0.3, current state is center=1.0, threshold=0.2;
    # we should send.
    my_meta.current_state = distant_state
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is True
    my_meta.sent_current_state()

    # Previous state sent and current state are the same; we should not send.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id_ignored, neighbors={}) is False


def test_geometric_term() -> None:
    '''Test the distance decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'self_distance': {
                    'threshold': 8.0,
                    'geometric_term': 0.5,
                    'min': 0.1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('self_distance')(
        config=distributed.decider.self_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    initial_state = ToyNeighborState(center=0.3)
    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=initial_state)
    my_meta.sent_current_state()
    my_meta.current_state = my_state

    neighbor_id = NodeID(-1)
    neighbors: Mapping[NodeID, NeighborMetadataInterface] = {}

    assert 'self_distance' not in my_meta.scratch
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 4.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 2.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 1.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 0.5
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 0.25
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['self_distance']['threshold'] == 0.125
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['self_distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['self_distance']['threshold'] == 0.1


def test_linear_term() -> None:
    '''Test the distance decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'self_distance': {
                    'threshold': 4.0,
                    'linear_term': -1.0,
                    'min': 0.1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('self_distance')(
        config=distributed.decider.self_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    initial_state = ToyNeighborState(center=0.3)
    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=initial_state)
    my_meta.sent_current_state()
    my_meta.current_state = my_state

    neighbor_id = NodeID(-1)
    neighbors: Mapping[NodeID, NeighborMetadataInterface] = {}

    assert 'self_distance' not in my_meta.scratch
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 3.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 2.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 1.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['self_distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['self_distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True



'''Provide threads that report exceptions back to the main thread.'''

from threading import Thread
from typing import Optional


class ExceptionThread(Thread):
    '''A thread that reports exceptions back to the main thread.'''
    _exc: Optional[BaseException] = None

    def run(self) -> None:
        try:
            super().run()
        except BaseException as e:  # pylint: disable=broad-exception-caught
            self._exc = e

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._exc:
            raise self._exc



'''Data loader for tests that ignores the data and returns an empty dataset.'''

from typing import Optional

from ..wrangler.dataset import Dataset

from .impl.dataframe_loader import DataframeLoader
from .impl.data_loader_catalog import DataLoaderCatalog

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class IgnoreDataLoader(DataframeLoader):
    '''Data loader for tests that ignores the data and returns a None dataset.'''
    name = 'ignore'
    tags = {
        'input_format': ['ignore'],
        'loaded_format': ['ignore'],
    }

    def _load_train(self) -> Optional[Dataset]:
        return None

    def _load_test(self) -> Optional[Dataset]:
        return None

    def poll(self, timeout: Optional[float] = 0) -> Optional[Dataset]:
        return None


def register(catalog: DataLoaderCatalog):
    '''Register all the objects in this file.'''
    catalog.register(IgnoreDataLoader)



'''Pipeline template for a tabular regression problem.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=too-many-arguments,duplicate-code
from typing import Dict, List, Optional

from ..algorithms.column_parser import ColumnParser
from ..algorithms.connect import ConnectorModel
from ..algorithms.extract_columns_by_role import ExtractColumnsByRoleModel
from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..algorithms.sklearn.impute.simple_imputer import SimpleImputerModel
from ..algorithms.sklearn.preprocessing.one_hot_encoder import OneHotModel
from ..problem_def.task import DataType, TaskType
from ..templates.impl.pipeline_step import GeneratorInterface
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog
from ..wrangler.dataset import DatasetKeys, RoleName


class TabularRegressionTemplate(PipelineTemplate):
    '''Simple implementation of a template for tabular regression.'''
    _tags: Dict[str, List[str]] = {
        'task': [TaskType.REGRESSION.name],
        'data_type': [DataType.TABULAR.name],
    }
    _name = 'tabular_regression'

    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 generator: Optional[GeneratorInterface] = None):
        super().__init__(name=name or self._name,
                         tags=tags or self._tags,
                         algorithm_catalog=algorithm_catalog, generator=generator)
        self.step(model=ColumnParser())

        attributes = self.new(name='attributes')
        attributes.step(
            model=ExtractColumnsByRoleModel(),
            desired_roles=RoleName.ATTRIBUTE)
        attributes.step(SimpleImputerModel())
        attributes.step(OneHotModel())

        target = self.new(name='target')
        target.step(
            model=ExtractColumnsByRoleModel(),
            desired_roles=RoleName.TARGET)

        self.parallel(target=target, attributes=attributes)
        self.step(
            model=ConnectorModel(),
            serialized_model=None,
            **{
                DatasetKeys.TARGET.value: [DatasetKeys.TARGET.value, DatasetKeys.DATAFRAME.value],
                DatasetKeys.COVARIATES.value: ['attributes', DatasetKeys.DATAFRAME.value]
            }
        )
        self.query(task=TaskType.REGRESSION.name).set_name('regressor')


def register(catalog: TemplateCatalog,
             algorithm_catalog: Optional[AlgorithmCatalog] = None,
             generator: Optional[GeneratorInterface] = None):
    '''Register all the objects in this file.'''
    catalog.register(
        TabularRegressionTemplate(
            algorithm_catalog=algorithm_catalog,
            generator=generator))



'''Representation of a neighbor for Auton Lab's distributed decision tree.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


import pickle
from typing import List, NamedTuple, Optional, Tuple

import numpy as np

from ...wrangler.logger import Logger, Level

from ..impl.distributed_algorithm_instance import NeighborState


log = Logger(__file__, to_stdout=False,
             level=Level.INFO).logger()


class Acorn(NamedTuple):
    '''Minimal information needed to represent a decision tree.

    See AutonDecisionTreeNeighbor for more info about what these mean.'''
    n_in: int
    n_out: int
    children_left: List[Optional[int]]
    children_right: List[Optional[int]]
    feature: List[Optional[int]]
    threshold: List[Optional[float]]
    equal: List[Optional[bool]]
    value: List[np.ndarray]
    sample_weight: List[float]
    lower_bound: np.ndarray
    upper_bound: np.ndarray
    class_list: List[str]


class Thresholds(NamedTuple):
    '''Threshold information.

    ts: length n_in.
        each element is a list of threshold candidates
    eqs: length n_in.
        each element is a list of bools indicating which threshold candidates are equal
    ddsdts: length n_in.
        each element has the same length as the corresponding element in ts and eqs
    '''
    ts: List[List[float]]
    eqs: List[List[bool]]
    ddsdts: List[List[float]]


class AutonDecisionTreeNeighbor(NeighborState):  # pylint: disable=too-many-public-methods
    '''State of a neighbor for a decision tree classifier.'''
    # TODO(Merritt/Kyle): rethink what NeighborState is (do we just want it to be what we send?)
    _acorn: Acorn

    def __init__(self,
                 acorn: Acorn,
                 columns: Optional[List[str]] = None):
        super().__init__(columns=columns)
        self._acorn = acorn

    @classmethod
    def init_tree(cls,
                  n_in: int,
                  n_out: int,
                  lower_bound: np.ndarray,
                  upper_bound: np.ndarray,
                  class_list: List[str],
                  columns: Optional[List[str]] = None
                  ) -> 'AutonDecisionTreeNeighbor':
        '''Create an empty tree to be built'''
        return cls(
            acorn=Acorn(
                n_in=n_in,
                n_out=n_out,
                children_left=[],
                children_right=[],
                feature=[],
                threshold=[],
                equal=[],
                value=[],
                sample_weight=[],
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                class_list=class_list),
            columns=columns)

    def __str__(self):
        return (f'{self.__class__.__name__}('
                f'lower_bound = {self.lower_bound}, \n'
                f'upper_bound = {self.upper_bound}, \n'
                f'class_list = {self.class_list}, \n'
                f'n_in = {self.n_in}, n_out = {self.n_out}, \n'
                f'number of nodes: {len(self.feature)}.\n')

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'acorn = {self._acorn})')

    def encode(self) -> bytes:
        '''Encode the state of the neighbor.'''
        retval = pickle.dumps((self._acorn, self._columns))
        return retval

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'AutonDecisionTreeNeighbor':
        '''Decode the state of the neighbor.'''
        (acorn, cols) = pickle.loads(serialized_model)
        assert isinstance(acorn, Acorn), (
            'BUG: expecting first element of message to be type Acorn, '
            f'instead found {acorn} of type {type(acorn)}'
        )
        return cls(columns=cls._cast_columns(cols), acorn=acorn)

    def norm2(self) -> float:
        '''Compute the norm of the state.'''
        return self.inner_product(self)

    def norm2_diff(self, other: 'AutonDecisionTreeNeighbor') -> float:
        '''Compute numerical representation of distance between models
        (||f-g||^2)
        '''
        self_norm = self.norm2()
        other_norm = other.norm2()
        inprod = self.inner_product(other)
        retval = self_norm + other_norm - 2.0 * inprod
        retval = round(retval, ndigits=5)  # Account for floating point error.
        assert retval >= 0.0, (
            f'BUG: norm2_diff should never be negative.  Got {retval}. '
            f'self.norm2() = {self_norm}; other.norm2() = {other_norm}, '
            f'self.inner_product(other) = {inprod}.'
        )
        return retval

    def distance(self, other: 'AutonDecisionTreeNeighbor') -> float:
        if self.class_list != other.class_list:
            return float('inf')

        return self.norm2_diff(other)

    def inner_product(self, other: 'AutonDecisionTreeNeighbor') -> float:
        '''Compute the inner product of two Decision trees.'''
        assert len(self.feature) > 0, 'BUG: taking inner product of an empty tree (self).'
        assert len(other.feature) > 0, 'BUG: taking inner product of an empty tree (other).'

        # this could be optimized, but probably doesn't need it
        # so this is the simple implementation
        lb = np.minimum(self.lower_bound, other.lower_bound)
        ub = np.maximum(self.upper_bound, other.upper_bound)
        feature_lengths = ub - lb
        # ignore all zero-length dimensions when computing the volume
        feature_lengths[feature_lengths == 0] = 1
        totl_vol = feature_lengths.prod()

        assert self.class_list == other.class_list, (
            'BUG: trying to take the inner product of decision trees with '
            'different class lists. \n'
            f'Self: {self.class_list} \n'
            f'Other: {other.class_list}'
        )

        def trav(i: int, lb: np.ndarray, ub: np.ndarray) -> float:
            if self.is_leaf(i):
                if other is self:
                    sub_feature_lengths = ub - lb
                    sub_feature_lengths[feature_lengths == 0] = 1
                    vol = sub_feature_lengths.prod()
                    retval = vol * np.sum(np.square(self.value[i]))
                    assert not np.isnan(retval)
                    return retval
                s = other.get_thresholds(lb, ub)[0]
                retval = s @ self.value[i]
                assert not np.isnan(retval)
                return retval
            f, t, _, left, right = self.internal_node_info(i)
            tot: float = 0.0
            old = ub[f]
            ub[f] = t
            tot += trav(left, lb, ub)
            ub[f] = old
            old = lb[f]
            lb[f] = t
            tot += trav(right, lb, ub)
            lb[f] = old
            return tot

        retval = trav(0, self.lower_bound.copy(), self.upper_bound.copy()) / totl_vol
        assert not np.isnan(retval)
        return retval

    def is_leaf(self, i: int) -> bool:
        '''Is node i a leaf?'''
        assert i < len(self._acorn.feature), (
            f'BUG: Asking if node {i} is a leaf but there are only '
            f'{len(self._acorn.feature)} nodes.')

        return self._acorn.feature[i] is None

    def internal_node_info(self, i: int) -> Tuple[int, float, bool, int, int]:
        '''Return (feature, threshold, equal, children_left, children_right)

        while asserting that this node is not a leaf.
        '''
        f = self.feature[i]
        t = self.threshold[i]
        eq = self.equal[i]
        left = self.children_left[i]
        right = self.children_right[i]
        assert (
            f is not None and t is not None and eq is not None
            and left is not None and right is not None), (
            'BUG: a non-leaf node must have non-None values for '
            f'feature (found {f}), threshold (for {t}), '
            f'equal (found {eq}), children_left (found {left}), '
            f'and children_right (found {right}).'
        )
        assert f < self.n_in, (
            f'BUG: feature {f} referenced with only {self.n_in} features.  '
            f'Full state: {repr(self)}'
        )
        return (f, t, eq, left, right)

    def get_thresholds(self,
                       lb: np.ndarray,
                       ub: np.ndarray
                       ) -> Tuple[np.ndarray, Thresholds]:
        '''return ss, ts, eqs, ddsdts

        ss: length n_out.
            something involving the sum of label weights for each class in a given region

        thresh: Threshold information.  Contains:
            ts: length n_in.
                each element is a list of threshold candidates
            eqs: length n_in.
                each element is a list of bools indicating which threshold candidates are equal
            ddsdts: length n_in.
                each element has the same length as the corresponding element in ts and eqs
        '''

        # initialize a list of outputs per input feature
        ss = np.zeros(self.n_out)

        thresh = Thresholds(
            ts=[[] for _ in range(self.n_in)],
            eqs=[[] for _ in range(self.n_in)],
            ddsdts=[[] for _ in range(self.n_in)]
        )

        if len(self.feature) == 0:
            # No threshold candidates as tree is empty
            log.warning('Calling get_thresholds on an empty tree: %s', repr(self))
            return ss, thresh

        # initialize bounds
        if self.lower_bound is not None:
            lb = np.maximum(lb, self.lower_bound)
        if self.upper_bound is not None:
            ub = np.minimum(ub, self.upper_bound)
        if np.any(lb >= ub):
            # No threshold candidates as domain is empty
            return ss, thresh

        def forward(i: int,  # pylint: disable=too-many-locals
                    lb: np.ndarray,
                    ub: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
            '''traverse tree

            i: index of tree-node
            lb: lower bound (n_in)
            ub: upper bound (n_in)

            returns:
                dsdt on the lower edge of the subtree (n_in x n_out)
                dsdt on the upper edge of the subtree (n_in x n_out)'''
            nonlocal ss, thresh
            if self.is_leaf(i):
                # leaves: compute dsdt for each edge
                w = ub - lb
                s: np.ndarray = self.value[i] * np.prod(w)
                ss += s
                dsdt = s[None, :] / w[:, None]
                return dsdt, dsdt
            # internal nodes: split left and right
            f, t, eq, child_l, child_r = self.internal_node_info(i)
            left = t > lb[f]
            right = t < ub[f]
            if left:
                old = ub[f]
                assert isinstance(old, float), f'Expected float, got type {type(old)} ({old})'
                ub[f] = min(old, t)
                dsdtl_l, dsdtu_l = forward(child_l, lb, ub)
                ub[f] = old
                if not right:
                    return dsdtl_l, dsdtu_l
            if right:
                old = lb[f]
                assert isinstance(old, float), f'Expected float, got type {type(old)} ({old})'
                lb[f] = max(old, t)
                dsdtl_r, dsdtu_r = forward(child_r, lb, ub)
                lb[f] = old
                if not left:
                    return dsdtl_r, dsdtu_r
            # append a threshold candidate for this internal node's split
            thresh.ts[f].append(t)
            thresh.eqs[f].append(eq)
            thresh.ddsdts[f].append(dsdtl_r[f] - dsdtu_l[f])
            # compute edge dsdt from child values
            dsdtl = dsdtl_l + dsdtl_r
            dsdtu = dsdtu_l + dsdtu_r
            dsdtl[f] -= dsdtl_r[f]
            dsdtu[f] -= dsdtu_l[f]
            return dsdtl, dsdtu

        # get thresholds from this tree's splits
        dsdtls, dsdtus = forward(0, lb, ub)

        # get thresholds for global bounds
        for t, eq, ddsdt, l, u, dsdtl, dsdtu in zip(
            thresh.ts, thresh.eqs, thresh.ddsdts, lb, ub, dsdtls, dsdtus
        ):
            t.append(l)
            eq.append(False)
            ddsdt.append(dsdtl)
            t.append(u)
            eq.append(True)
            ddsdt.append(-dsdtu)

        return ss, thresh

    def modify_classes(self, class_list: List[str]) -> None:
        '''Modify self in-place to have a new class list'''
        assert len(self.value[0]) == len(class_list), (
            'Mismatch between value and class list: '
            f'{len(self.value[0])} != {len(class_list)}'
        )
        self._acorn = Acorn(
            n_in=self.n_in,
            n_out=len(class_list),
            children_left=self.children_left,
            children_right=self.children_right,
            feature=self.feature,
            threshold=self.threshold,
            equal=self.equal,
            value=self.value,
            sample_weight=self.sample_weight,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            class_list=class_list
        )

    @property
    def n_in(self) -> int:
        '''Number of features'''
        return self._acorn.n_in

    @property
    def n_out(self) -> int:
        '''Number of classes'''
        return self._acorn.n_out

    @property
    def children_left(self) -> List[Optional[int]]:
        '''id of the left child of node i or None if leaf node.

        Length = number of nodes in the tree.
        '''
        return self._acorn.children_left

    @property
    def children_right(self) -> List[Optional[int]]:
        '''id of the right child of node i or None if leaf node.

        Length = number of nodes in the tree.
        '''
        return self._acorn.children_right

    @property
    def feature(self) -> List[Optional[int]]:
        '''feature used for splitting node i or None if leaf.

        Length = number of nodes in the tree.
        '''
        return self._acorn.feature

    @property
    def threshold(self) -> List[Optional[float]]:
        '''threshold value at node i or None if leaf.

        Length = number of nodes in the tree.
        '''
        return self._acorn.threshold

    @property
    def equal(self) -> List[Optional[bool]]:
        '''If true, decision threshold is <=; otherwise <  (None if leaf).'''
        return self._acorn.equal

    @property
    def value(self) -> List[np.ndarray]:
        '''Class probabilities at each leaf node.

        Each element is size n_out [not sure what it is at non-leafs].
        '''
        return self._acorn.value

    @property
    def sample_weight(self) -> List[float]:
        '''Representation of total amount of data present at each node in the tree.

        Length = number of nodes in the tree.
        '''
        return self._acorn.sample_weight

    @property
    def lower_bound(self) -> np.ndarray:
        '''Defines lower bound of the domain of the tree.

        Length n_in.
        '''
        return self._acorn.lower_bound

    @property
    def upper_bound(self) -> np.ndarray:
        '''Defines upper bound of the domain of the tree.

        Length n_in.
        '''
        return self._acorn.upper_bound

    @property
    def class_list(self) -> List[str]:
        '''List of observed classes in order.

        Length n_out.
        '''
        return self._acorn.class_list

    @property
    def num_nodes(self) -> int:
        '''Number of nodes in the tree.'''
        return len(self._acorn.threshold)


def register(catalog, *args, **kwargs) -> None:  # pylint: disable=unused-argument
    '''Nothing to register.'''



'''An Abstractor presents a view of the training space of neighbors.'''

from ...impl.algorithm import Algorithm

from .abstractor_instance import AbstractorInstance


class Abstractor(Algorithm):
    '''An abstractor algorithm.'''
    _name = 'abstractor'
    _default_hyperparams = {}

    def instantiate(self, **hyperparams) -> AbstractorInstance:
        raise NotImplementedError('instantiate is not implemented in Abstractor.')



'''Subset of NeighborState needed by deciders.'''

import abc
from typing import Generic, List, Optional, TypeVar

from ...config_components.distributed_config import DistributedConfig


NeighborStateSubclass = TypeVar('NeighborStateSubclass', bound='NeighborStateInterface')


class NeighborStateInterface(abc.ABC, Generic[NeighborStateSubclass]):
    '''Subset of NeighborState needed by deciders.'''
    @property
    @abc.abstractmethod
    def columns(self) -> Optional[List[str]]:
        '''The column names.'''

    @columns.setter
    @abc.abstractmethod
    def columns(self, value: List[str]) -> None:
        '''Set the column names.'''

    @abc.abstractmethod
    def state_differs(self,
                      distributed: DistributedConfig,
                      other: Optional[NeighborStateSubclass]) -> bool:
        '''Decide if our state differs meaningfully from another.

        :distributed: contains information from the problem definition, including fit_eps
        :other: is an instance of type(self) or None. It is often the last state we sent.
        '''

    @abc.abstractmethod
    def distance(self, other: NeighborStateSubclass) -> float:
        '''A numerical measure of the distance between this state and another.'''

    @property
    @abc.abstractmethod
    def payload_size(self) -> int:
        '''The size of the payload to send this state.'''

    @abc.abstractmethod
    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''

    @classmethod
    @abc.abstractmethod
    def decode(cls, serialized_model: bytes) -> NeighborStateSubclass:
        '''Decode a message from distributed neighbors.

        This should just redirect to the relevant NeighborState subclass.
        '''



'''Tests for cucumber.py'''
# pylint: disable=missing-class-docstring, missing-function-docstring,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ..algorithms.connect import ConnectorModel
from ..algorithms.impl.fake_algorithm import (FakeAlgorithm, FakeInstance,
                                              FAKE_SERIALIZED)
from ..generator.designator import Designator, StepDesignator

from .cucumber import Cucumber, DeserializationError


def test_cucumber_sunny_day() -> None:

    dut = Cucumber(
        impl=FakeAlgorithm().instantiate(),
        filename=StepDesignator('test_filename'),
        pipeline_designator=Designator('test_pipeline_designator'),
        hyperparams={'param 1': 'value_1'})

    got: Cucumber = Cucumber.deserialize(dut.serialize())

    assert got.filename == StepDesignator('test_filename')
    assert got.pipeline_designator == Designator('test_pipeline_designator')
    assert got.hyperparams['param 1'] == 'value_1'
    assert got.catalog_name == 'fake_algorithm'
    inst = got.deserialize_model(alg=FakeAlgorithm())
    assert isinstance(inst, FakeInstance)
    assert inst.evidence_of_deserialize == FAKE_SERIALIZED

    # We want deserialize_model() to create the instance with hyperparams saved from the cucumber,
    # instead of relying on the serialized model to hold its own hyperparams properly.
    assert inst.hyperparams() == {'param 1': 'value_1'}


def test_cucumber_deserialize_with_incorrect_alg() -> None:
    dut = Cucumber(
        impl=FakeAlgorithm().instantiate(),
        filename=StepDesignator('test_filename'),
        pipeline_designator=Designator('test_pipeline_designator'),
        hyperparams={})

    got: Cucumber = Cucumber.deserialize(dut.serialize())
    with pytest.raises(DeserializationError,
                       match=r'(connect.*fake_algorithm)|(fake_algorithm.*connect)'):
        got.deserialize_model(alg=ConnectorModel())
